from typing import Optional

import torch
from torch.nn import functional as F

from lumo.contrib.nn.functional import masked_log_softmax


def contrastive_loss(query: torch.Tensor, key: torch.Tensor,
                     norm=False,
                     temperature=0.7,
                     inbatch_neg=False,
                     graph_mask=None,
                     ):
    """

    Examples:
        from lumo.contrib.nn import functional as LF
        logits = torch.rand(4,128)
        logits_s = torch.rand(4,128)
        contrastive_loss(logits,logits_s,norm=True,inbatch_neg=True)

        logits = LF.normalize(logits)
        logits_s = LF.normalize(logits_s)
        contrastive_loss(logits,logits_s,norm=False,inbatch_neg=True)

        # sanity check
        # see `InstanceLoss` from https://github.com/Yunfan-Li/Contrastive-Clustering/blob/main/modules/contrastive_loss.py
        loss_fn = InstanceLoss(4,0.7,'cpu')
        loss_fn(logits,logits_s)
        contrastive_loss(logits,logits_s,inbatch_neg=True,temperature=0.7)

    Args:
        query: [Bq, feature_dim]
        key: [Bk, feature_dim]
        norm: bool, norm or not
        temperature: float, temperature
        inbatch_neg: bool,
        graph_mask: [Bq, Bk] or [Bq+Bk, Bq+Bk] if inbatch_neg = True

    Returns:

    """
    # if norm:
    #     query = normalize(query, dim=-1)
    #     key = normalize(key, dim=-1)

    qbs = len(query)

    if inbatch_neg:
        query = torch.cat([query, key])
        key = query
    if norm:
        logits = torch.cosine_similarity(query.unsqueeze(1), key.unsqueeze(0), dim=2)
    else:
        logits = torch.mm(query, key.t())  # query @ key.t()

    # apply temperature
    logits /= temperature

    # make label
    if graph_mask is None:
        graph_mask = torch.eye(len(logits), dtype=torch.float, device=query.device)
        if inbatch_neg:
            graph_mask = torch.cat([graph_mask[qbs:], graph_mask[:qbs]])

    # make same sample mask
    smask = None
    if inbatch_neg:
        smask = (1 - torch.eye(len(logits), dtype=torch.float, device=query.device))
        graph_mask = (graph_mask * smask)

    loss = -torch.sum(masked_log_softmax(logits, smask, dim=-1) * graph_mask, dim=1)
    loss = (loss / graph_mask.sum(1)).mean()  # when self-supervised without cluster, it equals to loss.mean()
    return loss


def contrastive_loss2(query: torch.Tensor, key: torch.Tensor,
                      memory: Optional[torch.Tensor] = None,
                      norm: Optional[bool] = False,
                      temperature: float = 0.7,
                      query_neg: Optional[bool] = False,
                      key_neg: Optional[bool] = True,
                      qk_graph: Optional[torch.Tensor] = None,
                      qm_graph: Optional[torch.Tensor] = None,
                      eye_one_in_qk: Optional[bool] = False,
                      softmax_mask=None,
                      reduction: Optional[str] = 'mean'
                      ):
    """
    Examples:
        >>> assert contrastive_loss(a,b,inbatch_neg=False) == contrastive_loss2(a,b,query_neg=False,key_neg=True)

    Args:
        query: [bq, f_dim]
        key: [bq, f_dim]
        memory: [bm, f_dim]
        norm: bool, normalize feature or not
        temperature: float
        query_neg: bool
        key_neg: bool
        qk_graph: [bq, bq], >= 0
        qm_graph: [bq, bm], >= 0
        eye_one_in_qk: bool
        reduction: str
            'mean',
            'sum',
            'none'

    Returns:
        loss

    """
    if memory is None:
        key_neg = True
        qm_graph = None

    if memory is not None:
        key = torch.cat([key, memory])

    if query_neg:
        key = torch.cat([query, key])

    q_size = query.shape[0]

    if norm:
        logits = torch.cosine_similarity(query.unsqueeze(1), key.unsqueeze(0), dim=2)
    else:
        logits = torch.mm(query, key.t())  # query @ key.t()

    logits /= temperature

    neg_index = torch.ones_like(logits, dtype=torch.bool, device=logits.device)

    _temp_eye_indice = torch.arange(q_size, device=logits.device).unsqueeze(1)

    neg_offset = q_size if query_neg else 0
    if query_neg:
        neg_index.scatter_(1, _temp_eye_indice, 0)

    if key_neg:
        neg_index.scatter_(1, _temp_eye_indice + neg_offset, 0)
    else:
        neg_index[:, neg_offset:neg_offset + q_size] = 0

    pos_index = torch.zeros_like(logits, dtype=torch.float, device=logits.device)
    pos_offset = q_size if query_neg else 0

    # for supervised cs
    if qk_graph is not None:
        pos_index[:, :q_size] = qk_graph.float()
        if query_neg:
            pos_index[:, q_size:q_size * 2] = qk_graph.float()

    # for supervised cs with moco memory bank
    if qm_graph is not None:
        pos_index[:, pos_offset + q_size:] = qm_graph.float()

    if qk_graph is None or eye_one_in_qk:
        pos_index.scatter_(1, _temp_eye_indice + pos_offset, 1)

    logits_mask = (pos_index > 0) | neg_index
    if softmax_mask is not None:
        if logits_mask.shape != softmax_mask.shape:
            raise ValueError(f'softmax_mask.shape should be {logits_mask.shape}, but got {softmax_mask.shape}')
        logits_mask = softmax_mask.bool() * logits_mask
    loss = -torch.sum(masked_log_softmax(logits, logits_mask, dim=-1) * pos_index, dim=1)
    loss = (loss / pos_index.sum(1))
    if reduction == 'mean':
        loss = loss.mean()
    elif reduction == 'sum':
        loss = loss.sum()

    return loss


def cluster_loss(query, key, label_graph_mask=None, temperature=0.7):
    """
    see https://github.com/Yunfan-Li/Contrastive-Clustering

    Examples:
        assert (prob_a.sum(dim=-1) == 1).all()
        assert (prob_b.sum(dim=-1) == 1).all()
        cluster_loss(prob_a,prob_b)

        # sanity check
        # see `ClusterLoss` from https://github.com/Yunfan-Li/Contrastive-Clustering/blob/main/modules/contrastive_loss.py
        loss_fn = ClusterLoss(5,0.7,'cpu')
        loss_fn(prob_a,prob_b)
        cluster_loss(prob_a,prob_b,temperature=0.7)

    """
    return contrastive_loss(query.t(), key.t(), norm=True,
                            temperature=temperature,
                            inbatch_neg=True,
                            graph_mask=label_graph_mask)


def sup_contrastive_loss(query: torch.Tensor, key: torch.Tensor,
                         label=None,
                         norm=False,
                         temperature=0.7):
    """
    A modified version from https://github.com/HobbitLong/SupContrast

    Examples:
        q = torch.rand(4,128)
        k = torch.rand(4,128)
        label = torch.randint(0,10,(4,))
        sup_contrastive_loss(q,k,label,norm=True)

        # sanity check
        sup = SupConLoss()
        sup(torch.stack([a,b],1))
        sup_contrastive_loss(a,b,norm=False,temperature=0.07)

    """
    if label is None:
        graph = None
    else:
        label = label.unsqueeze(0)
        graph = (label == label.T)
        graph = graph.repeat(2, 2)
    return contrastive_loss(
        query, key,
        norm=norm, temperature=temperature, inbatch_neg=True,
        graph_mask=graph,
    )


def cross_entropy_with_targets(logits, targets, mask=None):
    """

    Args:
        logits: [bs, class_num]
        targets: [bs, class_num]
        mask: [bs,]

    Returns:

    """
    out = torch.sum(F.log_softmax(logits, dim=-1) * targets, dim=1)
    if mask is not None:
        out = out * mask.float()
    return -torch.mean(out)


def minent(logits, w_ent=1):
    loss = - torch.sum(F.log_softmax(logits, dim=-1) * F.softmax(logits, dim=-1), dim=-1).mean() * w_ent
    return loss



from typing import Optional, List, Union

import torch
from accelerate.data_loader import (IterableDatasetShard, BatchSamplerShard, _PYTORCH_DATALOADER_KWARGS,
                                    DataLoaderDispatcher as _DataLoaderDispatcher, DataLoaderShard as _DataLoaderShard)
from accelerate.state import AcceleratorState, DistributedType, is_tpu_available
from accelerate.utils import RNGType, synchronize_rng_states, send_to_device
from torch.utils.data import DataLoader, IterableDataset
from .loader import LumoDataLoader, DataLoaderIterWrap

if is_tpu_available():
    import torch_xla.core.xla_model as xm


class DataLoaderDispatcher(_DataLoaderDispatcher, LumoDataLoader):

    def __iter__(self) -> DataLoaderIterWrap:
        return super().__iter__()


class DataLoaderShard(_DataLoaderShard, LumoDataLoader):

    def __iter__(self):
        if self.rng_types is not None:
            synchronize_rng_states(self.rng_types, self.generator)
        state = AcceleratorState()
        for batch in LumoDataLoader.__iter__(self):
            if state.distributed_type == DistributedType.TPU:
                xm.mark_step()
            yield batch if self.device is None else send_to_device(batch, self.device)


def prepare_data_loader(
        dataloader: DataLoader,
        device: Optional[torch.device] = None,
        num_processes: Optional[int] = None,
        process_index: Optional[int] = None,
        split_batches: bool = False,
        put_on_device: bool = False,
        rng_types: Optional[List[Union[str, RNGType]]] = None,
        dispatch_batches: Optional[bool] = None,
) -> DataLoader:
    """
    Wraps a PyTorch :obj:`DataLoader` to generate batches for one of the processes only.

    Depending on the value of the :obj:`drop_last` attribute of the :obj:`dataloader` passed, it will either stop the
    iteration at the first batch that would be too small / not present on all processes or loop with indices from the
    beginning.

    Args:
        dataloader (:obj:`torch.utils.data.dataloader.DataLoader`):
            The data loader to split across several devices.
        device (:obj:`torch.device`):
            The target device for the returned :obj:`DataLoader`.
        num_processes (:obj:`int`, `optional`):
            The number of processes running concurrently. Will default to the value given by
            :class:`~accelerate.AcceleratorState`.
        process_index (:obj:`int`, `optional`):
            The index of the current process. Will default to the value given by :class:`~accelerate.AcceleratorState`.
        split_batches (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether the resulting :obj:`DataLoader` should split the batches of the original data loader across devices
            or yield full batches (in which case it will yield batches starting at the :obj:`process_index`-th and
            advancing of :obj:`num_processes` batches at each iteration).

            Another way to see this is that the observed batch size will be the same as the initial :obj:`dataloader`
            if this option is set to :obj:`True`, the batch size of the initial :obj:`dataloader` multiplied by
            :obj:`num_processes` otherwise.

            Setting this option to :obj:`True` requires that the batch size of the :obj:`dataloader` is a round
            multiple of :obj:`batch_size`.
        put_on_device (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether or not to put the batches on :obj:`device` (only works if the batches are nested list, tuples or
            dictionaries of tensors).
        rng_types (list of :obj:`str` or :class:`~accelerate.utils.RNGType`):
            The list of random number generators to synchronize at the beginning of each iteration. Should be one or
            several of:

            - :obj:`"torch"`: the base torch random number generator
            - :obj:`"cuda"`: the CUDA random number generator (GPU only)
            - :obj:`"xla"`: the XLA random number generator (TPU only)
            - :obj:`"generator"`: the :obj:`torch.Generator` of the sampler (or batch sampler if there is no sampler in
              your dataloader) or of the iterable dataset (if it exists) if the underlying dataset is of that type.

        dispatch_batches (:obj:`bool`, `optional`):
            If set to :obj:`True`, the datalaoder prepared is only iterated through on the main process and then the
            batches are split and broadcast to each process. Will default to :obj:`True` when the underlying dataset is
            an :obj:`IterableDataset`, :obj:`False` otherwise.

    Returns:
        :obj:`torch.utils.data.dataloader.DataLoader`: A new data loader that will yield the portion of the batches

    .. warning::

        This does not support :obj:`BatchSampler` with varying batch size yet.
    """
    if dispatch_batches is None:
        dispatch_batches = False if not put_on_device else isinstance(dataloader.dataset, IterableDataset)

    if dispatch_batches and not put_on_device:
        raise ValueError("Using `dispatch_batches=True` requires `put_on_device=True`.")
    # Grab defaults from AcceleratorState
    state = AcceleratorState()
    if num_processes is None:
        num_processes = state.num_processes
    if process_index is None:
        process_index = state.process_index

    # Sanity check
    if split_batches and dataloader.batch_size > 1 and dataloader.batch_size % num_processes != 0:
        raise ValueError(
            f"To use a `DataLoader` in `split_batches` mode, the batch size ({dataloader.batch_size}) "
            f"needs to be a round multiple of the number of processes ({num_processes})."
        )

    new_dataset = dataloader.dataset
    # Iterable dataset doesn't like batch_sampler, but data_loader creates a default one for it
    new_batch_sampler = dataloader.batch_sampler if not isinstance(new_dataset, IterableDataset) else None
    generator = getattr(dataloader, "generator", None)
    # No change if no multiprocess
    if num_processes != 1 and not dispatch_batches:
        if isinstance(new_dataset, IterableDataset):
            if getattr(dataloader.dataset, "generator", None) is not None:
                generator = dataloader.dataset.generator
            new_dataset = IterableDatasetShard(
                new_dataset,
                batch_size=dataloader.batch_size,
                drop_last=dataloader.drop_last,
                num_processes=num_processes,
                process_index=process_index,
                split_batches=split_batches,
            )
        else:
            # New batch sampler for the current process.
            if hasattr(dataloader.sampler, "generator"):
                if dataloader.sampler.generator is None:
                    dataloader.sampler.generator = torch.Generator()
                    generator = dataloader.sampler.generator
                    generator.manual_seed(int(torch.empty((), dtype=torch.int64).random_().item()))
            elif getattr(dataloader.batch_sampler, "generator", None) is not None:
                generator = dataloader.batch_sampler.generator
            new_batch_sampler = BatchSamplerShard(
                dataloader.batch_sampler,
                num_processes=num_processes,
                process_index=process_index,
                split_batches=split_batches,
            )

    # We ignore all of those since they are all dealt with by our new_batch_sampler
    ignore_kwargs = [
        "batch_size",
        "shuffle",
        "sampler",
        "batch_sampler",
        "drop_last",
        "generator",
    ]

    if rng_types is not None and generator is None and "generator" in rng_types:
        rng_types.remove("generator")

    kwargs = {
        k: getattr(dataloader, k, _PYTORCH_DATALOADER_KWARGS[k])
        for k in _PYTORCH_DATALOADER_KWARGS
        if k not in ignore_kwargs
    }

    # Need to provide batch_size as batch_sampler is None for Iterable dataset
    if new_batch_sampler is None:
        kwargs["batch_size"] = dataloader.batch_size // num_processes if split_batches else dataloader.batch_size

    if dispatch_batches:
        res = DataLoaderDispatcher(
            new_dataset, split_batches=split_batches, batch_sampler=new_batch_sampler, **kwargs
        )
    else:
        res = DataLoaderShard(
            new_dataset,
            device=device if put_on_device else None,
            batch_sampler=new_batch_sampler,
            rng_types=rng_types,
            generator=generator,
            **kwargs,
        )

    if isinstance(dataloader, LumoDataLoader):
        res.set_prop(dataloader._prop)

    return res



import bisect
import sys
import warnings
from datetime import datetime
from functools import lru_cache
from typing import Union, Dict, Any, Optional, Sequence, Mapping, Callable

import numpy as np
import torch
from lumo.contrib.accelerate import Accelerator
from lumo.contrib.accelerate.utils import send_to_device
# overwrite send_to_device to resolve https://github.com/pytorch/pytorch/issues/83015
# from accelerate import Accelerator
# from accelerate.utils import send_to_device

from accelerate import DistributedDataParallelKwargs
from torch import nn
from torch.optim import Optimizer
from torch.utils.data import DataLoader

from lumo.core import ParamsType, TrainStage, Record, MetricType, Meter, Attr
from lumo.core.disk import TableRow, Metrics
from lumo.data import DataModule
from lumo.data.accelerator import DataLoaderShard, DataLoaderDispatcher
from lumo.data.loader import DataLoaderType, DataLoaderSide
from lumo.proc import dist
from lumo.trainer.rnd import RndManager
from lumo.utils.logger import Logger
from .base import _BaseTrainer
from .components import TrainerExperiment
from .saver import Saver


class Trainer(_BaseTrainer):
    """
    Trainer provides a complete training and evaluation logic to help users focus on specific training details.
    It integrates logging, metric recording, experiment saving, version control, and complete callback control
    of each life cycle.

    When to use lumo.Trainer:
     - At the beginning of the investigation direction, there will be many code branches and experiments
        based on the same set of data and models, but with different training details.
     - Reproduce multiple papers in the same field, which are also based on the same dataset and evaluation logic.

    When to simply use lumo.SimpleExperiment:
     - You have maintained a mature training framework yourself, but there are still certain requirements
        for allocating dynamic(different) storage path and version control in each program.

    At present, I have explored best practices based on lumo in two fields, this includs:
     - https://github.com/pytorch-lumo/image-classification
     - https://github.com/pytorch-lumo/MMERC

    Both two reposiroties are github templates and you can use them to create you new (private) repository.
    """

    callback_function = {
        "save_keypoint", "save_checkpoint", "save_model", "load_state_dict",
        'ioptims', 'imodels', 'train', 'test', 'evaluate', 'train_epoch', 'train_step',
        'test_step', 'evaluate_step', 'process_loader', 'regist_dataloader'
    }

    def __init__(self, params: ParamsType, dm: DataModule = None):
        if dm is None:
            dm = DataModule(params)
        else:
            if dm.params is None:
                dm.params = params

        self.dm = dm  # type: DataModule
        self.params = params
        self._logger = None
        self._saver = None

        self.params.iparams()
        self.exp = TrainerExperiment(self.generate_exp_name())

        self.database = TableRow(self.exp.project_name, self.exp.exp_name, self.exp.test_name_with_dist)
        self.metric_board = Metrics(self.exp.test_root)
        self.exp.dump_info('metric_board', self.metric_board.fpath)
        self.exp.dump_info('table_row', self.database.fpath)
        self.rnd = RndManager()

        self.train_epoch_toggle = False
        self.train_toggle = False

        device = params.get('device', None) if not self.is_dist else None

        self.accelerate = Accelerator(kwargs_handlers=[
            DistributedDataParallelKwargs(find_unused_parameters=params.get('find_unused_parameters', False))
        ])

        if self.accelerate.state.distributed_type == self.accelerate.state.distributed_type.NO:
            self.accelerate.state.device = torch.device(device)

        if dist.is_main():
            self.params.to_yaml(self.exp.test_file('params.yaml'))

        self.set_global_steps(0)
        self.set_epoch_idx(0)
        self.set_idx(0)
        if params.get('debug', False):
            self.exp.set_prop('debug', True)

    @property
    def metrics(self):
        return self.metric_board

    @property
    def db(self):
        return self.database

    @property
    def saver(self) -> Saver:
        if self._saver is None:
            self._saver = Saver(self.exp.saver_dir)
        return self._saver

    @property
    def is_debug(self):
        return self.params.get('debug', False)

    @property
    def is_main(self):
        return dist.is_main()

    @property
    def is_dist(self):
        return dist.is_dist()

    @property
    def local_rank(self):
        return dist.local_rank()

    @property
    def world_size(self):
        return dist.world_size()

    @property
    def wandb(self):
        import wandb
        wandb.init(project=self.exp.exp_name, name=self.exp.test_name)
        wandb.config = self.params.to_dict()
        return wandb

    @property
    def logger(self):
        if self._logger is None:
            from lumo.utils.logger import set_global_logger
            self._logger = Logger()
            set_global_logger(self._logger)
            if self.params.get('debug', False):
                self._logger.set_verbose(Logger.V_DEBUG)
                self._logger.debug('Enable debug log.')
            if self.is_main:
                fn = self._logger.add_log_dir(self.exp.log_dir)
                self.exp.dump_info('logger_args', {'log_dir': fn})
        return self._logger

    @property
    def safe_writer(self):
        """see trainer.writer"""
        try:
            import tensorflow as tf
            import tensorboard as tb
            tf.io.gfile = tb.compat.tensorflow_stub.io.gfile
        except ImportError:
            pass

        return self.writer

    @property
    @lru_cache()
    def writer(self):
        """
        Notes:
        ------
        When using add_embedding, there may raise some exceptions cased by version conflict, here is some solutions:

        1. `tensorflow_core._api.v1.io.gfile` or `tensorflow_core._api.v2.io.gfile` has no attribute `get_filesystem`
        first, try upgrade tensorboard and tensorflow as followed version:
            tensorboard==2.0.2
            tensorflow==2.0.0

        If you still have the same problem, use this code as a temporary solution:

            import tensorflow as tf
            import tensorboard as tb
            tf.io.gfile = tb.compat.tensorflow_stub.io.gfile

        Use `trainer.safe_writer` to get a writter with these code added inside thexp.

        > This solution is referred by https://github.com/pytorch/pytorch/issues/30966 .


        2. You may cause PermissionError like: [Errno 13] Permission denied: '/tmp/.tensorboard-info/pid-20281.info'.
        The solution is to set environment variable TMPDIR

            export TMPDIR=/tmp/$USER;
            mkdir -p $TMPDIR;
            tensorboard --logdir ...

        One line command version:
            export TMPDIR=/tmp/$USER; mkdir -p $TMPDIR; tensorboard --logdir ....

        > This solution is referred by https://github.com/tensorflow/tensorboard/issues/2010 .

        Returns:
            A SummaryWriter instance
        """
        try:
            from torch.utils.tensorboard import SummaryWriter
        except ModuleNotFoundError:
            try:
                from tensorboardX import SummaryWriter
            except ModuleNotFoundError:
                return None

        kwargs = self.exp.board_args
        res = SummaryWriter(**kwargs)

        def close(*args):
            res.flush()
            res.close()

        self.exp.add_exit_hook(close)
        return res

    @property
    def first_epoch(self):
        """Determine whether the current epoch is the first"""
        return self.eidx == 0

    @property
    def first_step(self):
        """Determine whether the current step is the first"""
        return self.global_steps == 0

    @property
    def idx(self):
        # started from 0
        return self._prop.get('idx', 0)

    @property
    def eidx(self):
        """
        This value will be automatically incremented by 1 before calling the train_epoch method
        current epoch round. The value remains unchanged in below methods:
         - `.on_train_epoch_begin()`: eidx
         - `.train_step()`: eidx
         - `.on_train_epoch_end()`: eidx

        This behaviour of `.idx` and `.global_steps` is the same as `.eidx`
        """
        return self._prop.get('eidx', 0)

    @property
    def global_steps(self) -> int:
        # started from 0
        return self._prop['global_steps']

    @property
    def trainer_state(self):
        return self._prop

    @property
    def devices(self) -> Dict[str, torch.device]:
        return self._state_dicts['devices']

    @property
    def model_dict(self) -> Dict[str, nn.Module]:
        return {key: self[key]
                for key in self._state_dicts['models']}

    @property
    def optim_dict(self) -> Dict[str, Optimizer]:
        return {key: self[key] for key in self._state_dicts['optims']}

    @property
    def torch_tensor(self) -> Dict[str, torch.Tensor]:
        return {key: self[key] for key in self._state_dicts['tensor.th']}

    @property
    def numpy_tensor(self) -> Dict[str, np.ndarray]:
        return {key: self[key] for key in self._state_dicts['tensor.np']}

    @property
    def others(self) -> Dict[str, Any]:
        return {key: self[key] for key in self._state_dicts['others']}

    @property
    def datamodule(self) -> DataModule:
        return self.dm

    @property
    def train_dataloader(self) -> Optional[DataLoaderType]:
        return self.datamodule['train']

    @property
    def test_dataloader(self) -> Optional[DataLoaderType]:
        return self.datamodule['test']

    @property
    def val_dataloader(self) -> Optional[DataLoaderType]:
        return self.datamodule['val']

    @property
    def device(self):
        return self.accelerate.device

    def _load_fun_state_dict(self, src: dict, tgt: dict):
        for k, v in tgt.items():
            if k in src:
                v.load_state_dict(src[k])

    def regist_dataloader(self, dataloader: DataLoader, stage: TrainStage):
        self.datamodule.regist_dataloader_with_stage(stage, dataloader)

    def process_loader(self, dm: Union[DataModule, DataLoader] = None, stage: TrainStage = TrainStage.train):
        """
        automatically called before train()/test()/evaluate(), see __new__ function of Trainer
        :param dm:
        :param stage:
        :return:
        """
        assert stage is not None, '`stage` cannot be None'
        if dm is None and self.dm is not None:
            dm = self.dm

        if isinstance(dm, DataModule):
            loader = dm[stage.value]
            if loader is None:
                # where datamodule.idataloader() methods first invoked (automaticly).
                loader = dm.get_loader_with_stage(stage)
                if loader is None:
                    return None
                loader = self.prepare_dataloader(loader, stage)
                self.regist_dataloader(loader, stage=stage)
        elif isinstance(dm, DataLoader):
            loader = dm
            loader = self.prepare_dataloader(loader, stage)
            self.regist_dataloader(loader, stage=stage)
        else:
            return None

        return loader

    def load_state_dict(self, state_dict: dict):
        _sub = {'models', 'optims', 'other'}
        _missing = []

        for k, v in state_dict.items():
            if k in _sub:
                self._load_fun_state_dict(v, self._state_dicts[k])
            else:
                self._state_dicts[k] = v
        return

    def to_device(self, item: Optional[Union[nn.Module, torch.Tensor, Sequence, Mapping]] = None,
                  device: torch.device = None):

        if item is None:
            for k, v in list(self.model_dict.items()):
                self[k] = self.accelerate.prepare(v)
            for k, v in list(self.optim_dict.items()):
                self[k] = self.accelerate.prepare(v)
        else:
            if device is None:
                device = self.device
            item = send_to_device(item, device)
            return item

    def on_trainer_exception(self, func: Callable, exception: BaseException):
        self.database.update_dict(dict(end=datetime.now(),
                                       finished=False,
                                       error=str(exception),
                                       trainer_frame=str(func)),
                                  flush=True)

    @property
    def is_initialized(self):
        if self._prop.get('initial', False):
            return True
        return False

    def initialize(self):
        if self.is_initialized:
            return
        self.exp.start()

        commit_info = self.exp.get_prop('git')
        commit_hex = None
        if commit_info is not None and 'commit' in commit_info:
            commit_hex = commit_info['commit']
        self.database.update('commit_hex', commit_hex)
        self.database.update_dict(dict(
            test_name=self.exp.test_name,
            path=self.exp.test_root,
            start=datetime.now()))
        self.database.set_params(self.params.to_dict())
        self.database.update('command', ' '.join(sys.argv))

        self.icallbacks(self.params)
        self.set_property('initial.callbacks', True)
        self.imodels(self.params)
        self.set_property('initial.model', True)
        self.set_property('initial', True)

    def stop_train(self):
        self.train_toggle = True
        self.train_epoch_toggle = True

    def stop_train_epoch(self):
        self.train_epoch_toggle = True

    def prepare_dataloader(self, loader: DataLoaderType, stage: TrainStage = None):
        """
        automatically called before train()/test()/evaluate(), see __new__ function of Trainer
        :param loader:
        :param stage:
        :return:
        """
        if isinstance(loader, (DataLoaderShard, DataLoaderDispatcher)):
            warnings.warn('Duplicated prepare a same DataLoader twice, check your code.')
            return loader

        split_batches = self.params.get('split_batches', None)
        if stage is not None and not stage.is_train():
            split_batches = True

        """do not change original loader stage"""
        if isinstance(loader, DataLoader):
            self.accelerate.split_batches = split_batches
            loader = self.accelerate.prepare_data_loader(loader)
        elif isinstance(loader, DataLoaderSide):
            loader = loader.copy()
            loader._loaders = {k: self.prepare_dataloader(v, stage) for k, v in loader._loaders.items()}
        return loader

    def train(self, dm: Union[DataModule, DataLoaderType] = None, params: ParamsType = None, limit_global_steps=None):
        loader = self.select_loader(dm)
        if not loader:
            loader = self.train_dataloader

        if loader is None:
            self.set_property('early_stop', 'Lack of train loader')
            return self._prop

        if params is None:
            params = self.params

        for eidx in range(params.epoch):
            self.set_epoch_idx(eidx)
            epoch_record = self.train_epoch(loader, params, limit_global_steps=limit_global_steps)
            self.set_property('record', epoch_record)
            self.set_property('record', epoch_record)
            if self.train_toggle:
                self.set_property('early_stop', 'train toggle')
                self.train_toggle = False
                break
            if limit_global_steps is not None and self.global_steps >= limit_global_steps:
                self.set_property('early_stop', f'meet limit_global_steps {limit_global_steps}')
                break

        self.database.update_dict(dict(end=datetime.now(), finished=True), flush=True)
        self.database.flush()
        return self._prop

    def train_epoch(self, loader: DataLoaderType, params: ParamsType = None,
                    limit_step=None,
                    limit_global_steps=None) -> Record:
        stage = TrainStage.train
        self.change_stage(stage)
        record = self.create_record(stage=stage)

        if params is None:
            params = self.params

        self.wait_for_everyone()
        for idx, batch in enumerate(loader):
            batch = send_to_device(batch, device=self.device)
            if self.train_epoch_toggle:
                self.train_epoch_toggle = False
                break

            if limit_step is not None and idx > limit_step:
                break
            if limit_global_steps is not None and self.global_steps >= limit_global_steps:
                break

            self.set_idx(idx)
            self._prop['global_steps'] += 1
            metric = self.train_step(batch, params)
            record.record(metric)
            self.database.flush()

        record.flush()
        self.database.update_dict(dict(eidx=self.eidx, end=datetime.now()))
        return record

    def set_property(self, key, value):
        self._prop[key] = value

    def set_global_steps(self, val):
        self.set_property('global_steps', val)

    def set_epoch_idx(self, val):
        self.set_property('eidx', val)

    def set_idx(self, val):
        self.set_property('idx', val)

    @property
    def trainstage(self) -> TrainStage:
        return self._prop.get('stage', TrainStage.default)

    def set_stage(self, val: TrainStage):
        self.set_property('stage', val)

    def add_callback(self, callback):
        """
        添加一个回调函数，注意，不能添加重复的 callback，这不推荐，也没有必要。
        :param callback:
        :return:
        """
        msg = None
        cb_name = callback.__class__.__name__
        cb_names = {cb.__class__.__name__ for cb in self.callbacks}
        if callback not in self.callbacks and cb_name in cb_names:
            msg = "Callback duplicate."
            callback.on_hook_failed(self, msg)
            return False

        if callback.only_main_process and not self.is_main:
            msg = f"{callback.__class__.__name__} only_main_process but in local_rank {self.local_rank}"
            callback.on_hook_failed(self, msg)
            return False

        if callback.only_single_gpu and not self.is_dist:
            msg = f"{callback.__class__.__name__} only_single_gpu but dist={self.is_dist}"
            callback.on_hook_failed(self, msg)
            return False

        if msg is not None:
            return False
        bisect.insort(self.callbacks, callback)

        callback._hooked = self
        callback.on_hooked(self, self.params)
        self.logger.info(f'{callback} hooked')
        return True

    def remove_callback(self, cur):
        self.callbacks.remove(cur)
        pass

    def change_stage(self, stage: TrainStage):
        if self.trainstage == stage:
            return

        self.set_stage(stage)
        for k, v in self.model_dict.items():
            if 'ema' in k.lower():
                continue
            if stage.value:
                v.train()
            else:
                v.eval()

    @classmethod
    def select_loader(cls, dm=None):
        loader = None
        if dm:
            if isinstance(dm, DataModule):
                loader = dm.train_dataloader
            elif isinstance(dm, DataLoader) or isinstance(dm, DataLoaderSide):
                loader = dm
            else:
                raise TypeError(type(dm))
        return loader

    def test(self, dm: Union[DataModule, DataLoader] = None, params: ParamsType = None, limit_step=None):
        stage = TrainStage.test
        self.change_stage(stage)

        loader = self.select_loader(dm)
        if not loader:
            loader = self.test_dataloader

        if loader is None:
            return None

        if params is None:
            params = self.params

        record = self.create_record(stage=stage)
        self.wait_for_everyone()
        for idx, batch in enumerate(loader):
            batch = send_to_device(batch, device=self.device)
            if limit_step is not None and idx >= limit_step:
                break
            self.set_idx(idx)
            metric = self.test_step(batch, params)
            record.record(metric)

        record.flush()
        return record

    def evaluate(self, dm: Union[DataModule, DataLoader] = None, params: ParamsType = None, limit_step: int = None):
        stage = TrainStage.val
        self.change_stage(stage)

        loader = self.select_loader(dm)
        if not loader:
            loader = self.val_dataloader
        if loader is None:
            return None

        if params is None:
            params = self.params

        record = self.create_record(stage=stage)
        for idx, batch in enumerate(loader):
            batch = send_to_device(batch, device=self.device)
            if limit_step is not None and idx >= limit_step:
                break
            self.set_idx(idx)
            metric = self.evaluate_step(batch, params)
            record.record(metric)

        record.flush()
        return record

    def train_step(self, batch, params: ParamsType = None) -> MetricType:
        pass

    def test_step(self, batch, params: ParamsType = None) -> MetricType:
        pass

    def evaluate_step(self, batch, params: ParamsType = None) -> MetricType:
        pass

    def imodels(self, params: ParamsType):
        pass

    def icallbacks(self, params: ParamsType):
        pass

    def inference(self, batch):
        raise NotImplementedError()

    def predict(self, batch):
        raise NotImplementedError()

    def optim_state_dict(self, wrap=True):
        res = {k: v.state_dict() for k, v in self.optim_dict.items()}
        if wrap:
            res = {'optim': res}
        return res

    def model_state_dict(self, wrap=True):
        res = {k: self.accelerate.unwrap_model(v).state_dict() for k, v in self.model_dict.items()}
        if wrap:
            res = {'model': res}
        return res

    def other_state_dict(self, wrap=True):
        res = {k: v.state_dict() for k, v in self.others.items()}
        if wrap:
            res = {'other': res}
        return res

    def state_dict(self):
        res = {
            'optims': self.optim_state_dict(wrap=False),
            'models': self.model_state_dict(wrap=False),
            'others': self.other_state_dict(wrap=False),
            'thtensor': self.torch_tensor,
            'nptensor': self.numpy_tensor,
        }

        return res

    def Meter(self):
        return Meter()

    def create_record(self, stage: TrainStage = None):
        if stage is None:
            stage = self.trainstage
        record = Record(stage=stage)
        return record

    def wait_for_everyone(self):
        """
        making sure all processes have reached this point before continuing.
        """
        self.accelerate.wait_for_everyone()

    def save_model(self, is_best=False, meta_info: Union[str, dict, Attr] = None):
        info = self._build_trainer_meta_info(meta_info)
        val = self.saver.save_model(self.eidx, self.model_state_dict(),
                                    meta_info=info,
                                    is_best=is_best)
        self.wait_for_everyone()
        return val

    def _build_trainer_meta_info(self, meta_info: Union[str, dict, Attr] = None):
        info = dict()
        info['eidx'] = self.eidx
        if meta_info is not None:
            if isinstance(meta_info, str):
                info['msg'] = meta_info
            if isinstance(meta_info, Meter):
                meta_info = meta_info.serialize()
            if isinstance(meta_info, dict):
                info.update(meta_info)
        return info

    def save_checkpoint(self, max_keep=10, is_best=False, meta_info: Union[str, dict, Meter] = None):
        info = self._build_trainer_meta_info(meta_info)
        val = self.saver.save_checkpoint(self.eidx, self.state_dict(),
                                         meta_info=info,
                                         max_keep=max_keep,
                                         is_best=is_best)
        self.wait_for_everyone()
        return val



"""
"""
import inspect
import json
import os
import tempfile
import time
import traceback
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from functools import wraps
from typing import NewType, Any, Optional, Dict, Union

from torch.utils.data import DataLoader

from lumo.core import ParamsType, Meter, MetricType, Record, TrainStage, wrap_result
from lumo.data import DataModule
from lumo.data.loader import summarize_loader, DataLoaderType
from lumo.utils import fmt
from lumo.utils.screen import inlinetqdm
from .trainer import Trainer
from ..proc.dist import world_size

_func_map = {
    'evaluate': 'eval',
    'evaluate_step': 'eval_step'
}


def map_func_name(name):
    return _func_map.get(name, name)


class BaseCallback:
    """
    Abstract base class used to build new callbacks.

    Callbacks are created to hook into the various stages of lumo's Trainer training, inference or data preparing.
    You can hook any subclasses of `BaseCallback` in `lumo.kit.trainer.Trainer` class.

    To create a custom callback class, subclass `lumo.kit.callbacks.BaseCallback`, and define the callback methods
    you want to do something before or after that.

    The callback method must be named like on_xxx_begin()/on_xxx_end(), where `xxx` is a function in trainer.


    TrainCallback, SaveLoadCallback and InitialCallback define the most methods in Trainer, you can simply subclass these
    callback class to custom your callback.

    Notes:
        There is no need for you to call the on_xxx function in your function's logic.

        The Trainer has some magic feature... not like keras or other framework with trainer/callback logic that
        you may need to manully call the callback function `on_xxx...()` in you overrided or newly defined method in Trainer,
        If you define a function `foo()` in your trainer, the callback function on_foo_begin() and on_foo_end() will be found
        automatically and be called, There is no need for you to call them in your function's logic.


    Examples:
        {{examples/docs/basecallback.py}}

        {{examples/docs/basecallback2.py}}

    """
    priority = 0  # type: int # All callbacks in lumo will have priority in range 0-100
    only_single_gpu = False  # Callback only works when single gpu mode
    only_main_process = False  # Callback only works when in main process

    def __new__(cls, *_, **__):
        self = super().__new__(cls)
        self._hooked = None  # type: Trainer

        def ecp_wrap(func):
            """Wrapper for cache exceptions"""

            @wraps(func)
            def on_exception(hooked: Trainer, tfunc, params: ParamsType, e: BaseException, *args, **kwargs):
                self.ecp = getattr(self, "ecp", None)
                res = None
                if self.ecp != e:
                    res = self.on_first_exception(hooked, tfunc, params, e, *args, **kwargs)
                    self.ecp = e

                eres = func(hooked, tfunc, params, e, *args, **kwargs)
                if res is None:
                    return eres
                else:
                    return res

            return on_exception

        self.on_exception = ecp_wrap(self.on_exception)
        return self

    def getfuncargs(self, func, *args, **kwargs):
        res = inspect.getcallargs(func, *args, **kwargs)
        if 'self' in res:
            res.pop('self')
        return res

    def on_hooked(self, source: Trainer, params: ParamsType):
        """called when callback hooked trainer"""
        pass

    def on_first_exception(self, source: Trainer, func, params: ParamsType, e: BaseException, *args, **kwargs):
        """
        when an exception was raised in some function, on_exception() will be called.

        如果异常发生在一个嵌套调用的函数中，那么该异常会在每一层函数都raise一次。

        该方法将被调用当该异常第一次raise出来的时候。
        该方法在 __new__ 中作了处理逻辑，不受继承关系影响
        """
        pass

    def on_exception(self, source: Trainer, func, params: ParamsType, e: BaseException, *args, **kwargs):
        """called when exception raised in some function"""
        return False

    def on_hook_failed(self, source, message):
        """Any reason when callback cannot hook on trainer"""
        pass

    def on_begin(self, source: Trainer, func, params: ParamsType, *args, **kwargs):
        """called before trainer.func is called"""
        cb_func = getattr(self, f"on_{map_func_name(func.__name__)}_begin", None)
        if cb_func is not None:
            cb_func(source, func, params, *args, **kwargs)

    def on_end(self, source: Trainer, func, params: ParamsType, result, *args, **kwargs):
        cb_func = getattr(self, f"on_{map_func_name(func.__name__)}_end", None)
        if cb_func is not None:
            cb_func(source, func, params, result, *args, **kwargs)

    def __le__(self, other):
        return self.priority <= other.priority

    def __lt__(self, other):
        return self.priority < other.priority

    def hook(self, source: Trainer):
        source.add_callback(self)

    def unhook(self):
        self._hooked.remove_callback(self)

    def _repr_by_val(self, *vals):
        vstr = "; ".join(["{}={}".format(val, str(getattr(self, val, None))) for val in vals])
        return "<{}([{}]) at 0x{:016X}>".format(self.__class__.__name__, vstr, id(self))

    def __repr__(self) -> str:
        return self._repr_by_val("priority")


class TrainCallback(BaseCallback):
    """
    Abstract base class used to build new callbacks. Defined some methods used in train/eval/test loop.

    Subclass TrainCallback if you want to control or add behavior before or after
        - train/train_epoch/train_step
        - eval/eval_epoch/eval_step
        - test/test_epoch/test_step
        - predict
        - inference
    """

    def on_train_begin(self, trainer: Trainer, func, params: ParamsType,
                       dm: Union[DataModule, DataLoaderType] = None, arg_params: ParamsType = None,
                       *args, **kwargs):
        pass

    def on_train_epoch_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        pass

    def on_test_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        pass

    def on_eval_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        pass

    def on_train_step_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        pass

    def on_eval_step_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        pass

    def on_test_step_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        pass

    def on_train_end(self, trainer: Trainer, func, params: ParamsType, record: Record, *args, **kwargs):
        pass

    def on_train_epoch_end(self, trainer: Trainer, func, params: ParamsType, record: Record, *args,
                           **kwargs):
        pass

    def on_test_end(self, trainer: Trainer, func, params: ParamsType, record: Record = None, *args, **kwargs):
        pass

    def on_eval_end(self, trainer: Trainer, func, params: ParamsType, record: Record = None, *args, **kwargs):
        pass

    def on_train_step_end(self, trainer: Trainer, func, params: ParamsType, metric: MetricType = None, *args,
                          **kwargs):
        pass

    def on_eval_step_end(self, trainer: Trainer, func, params: ParamsType, metric: MetricType = None, *args,
                         **kwargs):
        pass

    def on_test_step_end(self, trainer: Trainer, func, params: ParamsType, metric: MetricType = None, *args,
                         **kwargs):
        pass

    def on_predict_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        pass

    def on_predict_end(self, trainer: Trainer, func, params: ParamsType, result: Any = None, *args, **kwargs):
        pass

    def on_inference_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        pass

    def on_inference_end(self, trainer: Trainer, func, params: ParamsType, result: Any, *args, **kwargs):
        pass


class InitialCallback(BaseCallback):
    """
    Abstract base class used to build new callbacks. Defined some methods used in save/load.

    Subclass InitialCallback if you want to control or add behavior before or after
        - initialization of optims/models/dataloader
    """

    def on_imodels_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        pass

    def on_imodels_end(self, trainer: Trainer, func, params: ParamsType, result: Any, *args, **kwargs):
        pass

    def on_prepare_dataloader_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        pass

    def on_prepare_dataloader_end(self, trainer: Trainer, func, params: ParamsType, loader: DataLoader,
                                  *args, **kwargs):
        pass

    def on_process_loader_begin(self, trainer: Trainer, func, params: ParamsType,
                                dm: DataModule, stage: TrainStage,
                                *args, **kwargs):
        pass

    def on_process_loader_end(self, trainer: Trainer, func, params: ParamsType, loader: DataLoader,
                              dm: DataModule, stage: TrainStage,
                              *args, **kwargs):
        pass

    def on_regist_dataloader_begin(self, trainer: Trainer, func, params: ParamsType,
                                   dataloader: DataLoader, stage: TrainStage, *args, **kwargs):
        pass

    def on_regist_dataloader_end(self, trainer: Trainer, func, params: ParamsType, result: Any,
                                 dataloader: DataLoader, stage: TrainStage, *args, **kwargs):
        pass


class EvalCallback(TrainCallback):
    """
    Callback for evaluation.

    Args:
        eval_per_epoch:
        test_per_epoch:
    """

    def __init__(self, eval_per_epoch=1, test_per_epoch=10):

        self.eval_in_per_epoch = eval_per_epoch
        self.test_in_per_epoch = test_per_epoch

        # evaluate/test on train end
        self._last_eval = -1
        self._last_test = -1

    @property
    def has_eval(self):
        return self.eval_in_per_epoch is not None and self.eval_in_per_epoch > 0

    @property
    def has_test(self):
        return self.test_in_per_epoch is not None and self.test_in_per_epoch > 0

    def _test_or_eval(self, params: ParamsType, trainer: Trainer):
        if self.has_eval:
            if trainer.eidx % self.eval_in_per_epoch == self.eval_in_per_epoch - 1:
                self._last_eval = trainer.eidx
                trainer.evaluate()
        if self.has_test:
            if trainer.eidx % self.test_in_per_epoch == self.test_in_per_epoch - 1:
                self._last_test = trainer.eidx
                trainer.test()

    def on_train_epoch_end(self, trainer: Trainer, func, params: ParamsType, record: Record, *args,
                           **kwargs):
        self._test_or_eval(params, trainer)

    def on_train_end(self, trainer: Trainer, func, params: ParamsType, record: Record, *args, **kwargs):
        if self._last_eval != trainer.eidx and self.has_eval:
            trainer.evaluate()
        if self._last_test != trainer.eidx and self.has_test:
            trainer.test()

    def __repr__(self):
        return self._repr_by_val("eval_in_per_epoch", "test_in_per_epoch")


class DebugCallback(BaseCallback):

    def on_hooked(self, source: Trainer, params: ParamsType):
        super().on_hooked(source, params)
        print('on_hooked()')

    def on_first_exception(self, source: Trainer, func, params: ParamsType, e: BaseException, *args, **kwargs):
        super().on_first_exception(source, func, params, e, *args, **kwargs)
        print('on_first_exception()', func.__name__)

    def on_exception(self, source: Trainer, func, params: ParamsType, e: BaseException, *args, **kwargs):
        return super().on_exception(source, func, params, e, *args, **kwargs)

    def on_hook_failed(self, source, message):
        super().on_hook_failed(source, message)
        print('on_hook_failed()', message)

    def on_begin(self, source: Trainer, func, params: ParamsType, *args, **kwargs):
        super().on_begin(source, func, params, *args, **kwargs)
        print('on_begin()', func.__name__)

    def on_end(self, source: Trainer, func, params: ParamsType, result, *args, **kwargs):
        super().on_end(source, func, params, result, *args, **kwargs)
        print('on_end()', func.__name__)


class LoggerCallback(TrainCallback, InitialCallback):
    priority = 99999

    def __init__(self, step_frequence=3, break_in=1000):
        self.stage = {}
        self.breakin = break_in
        self.c = 0
        self.step = step_frequence
        file = tempfile.TemporaryFile('w')
        self.temp = file

    def on_imodels_end(self, trainer: Trainer, func, params: ParamsType, result: Any, *args, **kwargs):
        super().on_imodels_end(trainer, func, params, result, *args, **kwargs)
        trainer.logger.info('Model initialized.')

    def on_hooked(self, source: Trainer, params: ParamsType):
        super().on_hooked(source, params)
        source.logger.raw(params)

    def on_process_loader_end(self, trainer: Trainer, func, params: ParamsType, loader: DataLoaderType,
                              dm: DataModule,
                              stage: TrainStage, *args, **kwargs):
        super().on_process_loader_end(trainer, func, params, loader, dm, stage, *args, **kwargs)
        if loader is None:
            return
        if stage in self.stage:
            return
        loader_str = summarize_loader(loader)
        try:
            lsize = len(loader)
        except:
            lsize = None
        self.stage[stage] = lsize
        trainer.logger.info(f'{loader_str} for {stage.value} prepared.')

    def on_train_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        super().on_train_begin(trainer, func, params, *args, **kwargs)
        trainer.logger.raw('[[Train Begin]]')
        self.global_tqdm = inlinetqdm(total=params.epoch,
                                      position=0, leave=True,
                                      bar_format='Ela: {elapsed} | Last: {remaining} | Avg: {rate}',
                                      file=self.temp)

    def renew(self, stage):
        """创建一个新的"""
        self.cur_tqdm = inlinetqdm(total=self.stage[stage], position=0, leave=True,
                                   bar_format='{desc}{elapsed}<{remaining} ({percentage:3.0f}%){postfix}',
                                   file=self.temp)
        self.record = Record()

    def update(self, trainer: Trainer):
        self.c += 1
        self.cur_tqdm.update()
        if self.c % self.step == 0:
            trainer.logger.raw(self.cur_tqdm, inline=True)

        if self.c % self.breakin == 0 or (
                TrainStage.train in self.stage and ((trainer.idx + 1) == self.stage[TrainStage.train])):
            trainer.logger.info(self.cur_tqdm.full_str())
            # trainer.logger.newline()

    def flush(self, trainer: Trainer):
        self.c = 0
        trainer.logger.inline(self.cur_tqdm)
        trainer.logger.newline()

    def on_train_epoch_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        super().on_train_epoch_begin(trainer, func, params, *args, **kwargs)
        self.renew(TrainStage.train)
        self.time = time.time()

    def on_test_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        super().on_test_begin(trainer, func, params, *args, **kwargs)
        self.renew(TrainStage.test)
        trainer.logger.raw('[[Test Begin]]')

    def on_eval_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        super().on_eval_begin(trainer, func, params, *args, **kwargs)
        self.renew(TrainStage.val)
        trainer.logger.raw('[[Evaluate Begin]]')

    @staticmethod
    def format_interval(t: float):
        """
        Formats a number of seconds as a clock time, [H:]MM:SS

        Parameters
        ----------
        t  : int
            Number of seconds.

        Returns
        -------
        out  : str
            [H:]MM:SS
        """
        mins, s = divmod(int(t), 60)
        h, m = divmod(mins, 60)
        if h:
            return '{0:d}h{1:02d}m{2:02d}s'.format(h, m, s)
        else:
            return '{0:02d}m{1:02d}s'.format(m, s)

    def format_train_epoch_time(self, n, total, elapsed, ncols=None, prefix='', ascii=False, unit='it',
                                unit_scale=False, rate=None, bar_format=None, postfix=None,
                                unit_divisor=1000, initial=0, colour=None, **extra_kwargs):
        elapsed_str = self.format_interval(elapsed)
        remaining = (total - n) / rate if rate and total else 0
        remaining_str = self.format_interval(remaining) if rate else '?'

        if rate is None:
            rate_str = '?'
        else:
            rate_str = self.format_interval(1 / rate)

        last = time.time() - self.time
        last_str = self.format_interval(last)

        return f'[Epoch {n}/{total} End] cost: {elapsed_str} | current: {last_str} | remain: {remaining_str} | avg: {rate_str}'

    def on_train_epoch_end(self, trainer: Trainer, func, params: ParamsType, record: Record, *args, **kwargs):
        super().on_train_epoch_end(trainer, func, params, record, *args, **kwargs)
        self.global_tqdm.update()
        # self.flush(trainer)
        trainer.logger.raw(self.format_train_epoch_time(**self.global_tqdm.format_dict))

    def on_test_end(self, trainer: Trainer, func, params: ParamsType,
                    record: Optional[Record] = None,
                    *args, **kwargs):
        self.flush(trainer)
        trainer.logger.raw('[[Test End]]')

    def on_eval_end(self, trainer: Trainer, func, params: ParamsType,
                    record: Optional[Record] = None,
                    *args, **kwargs):
        self.flush(trainer)
        trainer.logger.raw('[[Evaluate End]]')

    def on_train_end(self, trainer: Trainer, func, params: ParamsType, record: Optional[Record], *args, **kwargs):
        self.flush(trainer)
        trainer.logger.raw('[[Train End]]')

    def on_train_step_end(self, trainer: Trainer, func, params: ParamsType,
                          metric: MetricType = None,
                          *args, **kwargs):
        self.record.record(metric)
        self.cur_tqdm.set_description_str(f"{trainer.idx + 1}/{self.stage[TrainStage.train]}, ", refresh=False)
        self.cur_tqdm.set_postfix_str(self.record.tostr(), refresh=False)
        self.update(trainer)

    def on_eval_step_end(self, trainer: Trainer, func, params: ParamsType,
                         metric: MetricType = None,
                         *args, **kwargs):
        self.record.record(metric)
        self.cur_tqdm.set_description_str(f"{trainer.idx + 1}/{self.stage[TrainStage.val]}, ", refresh=False)
        self.cur_tqdm.set_postfix_str(self.record.tostr(), refresh=False)
        self.update(trainer)

    def on_test_step_end(self, trainer: Trainer, func, params: ParamsType,
                         metric: MetricType = None,
                         *args, **kwargs):
        self.record.record(metric)
        self.cur_tqdm.set_description_str(f"{trainer.idx + 1}/{self.stage[TrainStage.test]}, ", refresh=False)
        self.cur_tqdm.set_postfix_str(self.record.tostr(), refresh=False)
        self.update(trainer)


class EpochCheckpoint(TrainCallback):
    """
    在 Trainer 训练过程中定时保存模型
    """
    only_main_process = True

    def __init__(self, per_epoch=50):
        self.per_epoch = per_epoch

    def on_train_epoch_end(self, trainer: Trainer, func, params: ParamsType, record: Optional[Record], *args,
                           **kwargs):
        meter = record.avg()
        if trainer.eidx % self.per_epoch == 0 and trainer.eidx > 0:
            trainer.save_checkpoint(meta_info=Meter.wrap_result(meter))

    def __repr__(self) -> str:
        return self._repr_by_val("per_epoch")


class GlobalStepCheckpoint(TrainCallback):
    only_main_process = True

    def __init__(self, per_step=2500):
        self.per = per_step

    def on_train_step_end(self, trainer: Trainer, func, params: ParamsType, metric: Meter, *args, **kwargs):
        super().on_train_step_end(trainer, func, params, metric, *args, **kwargs)
        if trainer.global_steps % self.per == 0 and trainer.global_steps > 0:
            trainer.save_checkpoint(meta_info=Meter.wrap_result(metric))


class KeyErrorSave(TrainCallback):
    """
    Callback to save checkpoints when you interrupt the program.
    """
    only_main_process = True
    only_single_gpu = True
    priority = -1

    def __init__(self, wait_input=False):
        self.wait_input = wait_input

    def on_first_exception(self, source: Trainer, func, params: ParamsType, e: BaseException, *args, **kwargs):
        if isinstance(e, KeyboardInterrupt):
            source.logger.info("KeyErrorSave trigged, save checkpoint")
            source.save_checkpoint({"mode": "KeyboardInterrupt"})

            tp = "n"
            if self.wait_input:
                tp = input("continue train step? (y/other)")

            if tp.lower() == "y":
                return True


class EMAUpdate(TrainCallback):
    """
    Callback to update EMA model every train step.

    Variable in Trainer instance is identified as EMA model when
     - is instance of torch.nn.Module
     - name is started with 'ema'
    """

    def on_train_step_end(self, trainer: Trainer, func, params: ParamsType, metric: MetricType, *args, **kwargs):
        super().on_train_step_end(trainer, func, params, metric, *args, **kwargs)
        for k, v in trainer.model_dict.items():
            if k.lower().startswith('ema'):
                v.step()


class AutoLoadModel(InitialCallback):
    """
    Callback to automatically load pretrained model.

    The load function will be executed if and only if
     - both the params `pretrain` and `pretrain_path` are defined in ParamsType
     - `pretrain` is True and `pretrain_path` is not None.
    """

    def on_imodels_end(self, trainer: Trainer, func, params: ParamsType, result: Any, *args, **kwargs):
        super().on_imodels_end(trainer, func, params, result, *args, **kwargs)
        if params.get('pretrain', False):
            path = params.get('pretrain_path', None)
            if path is not None:
                trainer.load_state_dict(path)


class EvalFirst(AutoLoadModel):
    """
    Callback to evaluation before the first train step. Aften used to debug or evaluate some dataset by restored model.
    """

    def __init__(self, datamodule=None):
        super().__init__()
        self.datamodule = datamodule

    def on_imodels_end(self, trainer: Trainer, func, params: ParamsType, result: Meter, *args, **kwargs):
        if params.get('eval_first', True):
            if self.datamodule is not None:
                trainer.evaluate(self.datamodule)
            else:
                trainer.evaluate()


class RecordCallback(TrainCallback):
    only_main_process = True

    def __init__(self, metric_step=150) -> None:
        super().__init__()
        self.metric_step = metric_step
        self.c = 0

    def log(self, metrics: MetricType, step, namespace):
        self.c += 1
        if self.c % self.metric_step != 0:
            return

        metrics = wrap_result(metrics)
        scalar_metrics = {}
        matrix_metrics = {}
        text_metrics = {}
        for k, v in metrics.items():
            v = fmt.to_ndarray(v)
            dtype = v.dtype.name
            elnum = v.size
            if 'str' in dtype:
                text_metrics[k] = str(v.tolist())
            else:
                if elnum == 1:
                    scalar_metrics[k] = v
                elif elnum > 1:
                    matrix_metrics[k] = v
        self.log_text(text_metrics, step, namespace)
        self.log_scalars(scalar_metrics, step, namespace)
        self.log_matrix(matrix_metrics, step, namespace)

    def log_text(self, metrics: Dict, step: int, namespace: str):
        return NotImplemented

    def log_scalars(self, metrics: Dict, step: int, namespace: str):
        return NotImplemented

    def log_matrix(self, metrics: Dict, step: int, namespace: str):
        return NotImplemented

    def on_hooked(self, source: Trainer, params: ParamsType):
        super().on_hooked(source, params)
        source.exp.set_prop('AutoRecord', self.__class__.__name__)

    def on_train_step_end(self, trainer: Trainer, func, params: ParamsType, metric: MetricType, *args, **kwargs):
        super().on_train_step_end(trainer, func, params, metric, *args, **kwargs)
        self.log(metric, step=trainer.global_steps, namespace='train.step')

    def on_train_epoch_end(self, trainer: Trainer, func, params: ParamsType, record: Record, *args, **kwargs):
        super().on_train_epoch_end(trainer, func, params, record, *args, **kwargs)
        self.log(record.avg(), step=trainer.global_steps, namespace='train.epoch')

    def on_test_end(self, trainer: Trainer, func, params: ParamsType, record: Record, *args, **kwargs):
        super().on_test_end(trainer, func, params, record, *args, **kwargs)
        self.log(record.avg(), step=trainer.global_steps, namespace='test')

    def on_eval_end(self, trainer: Trainer, func, params: ParamsType, record: Record, *args, **kwargs):
        super().on_eval_end(trainer, func, params, record, *args, **kwargs)
        self.log(record.avg(), step=trainer.global_steps, namespace='evaluate')


class WandbCallback(RecordCallback):
    only_main_process = True

    def __init__(self, metric_step=500) -> None:
        super().__init__()
        self.metric_step = metric_step
        self.c = 0

    def log(self, metrics: MetricType, step, namespace):
        self.c += 1
        if self.c % self.metric_step == 0:
            metrics = {
                f"{namespace}.{k}": v
                for k, v in wrap_result(metrics).items()}
            self._hooked.wandb.log(metrics, step=step)

    def log_text(self, metrics: Dict, step: int, namespace: str):
        wandb = self._hooked.wandb
        metrics = {k: v for k, v in metrics.items()}
        wandb.log(metrics, step=step)

    def log_scalars(self, metrics: Dict, step: int, namespace: str):
        wandb = self._hooked.wandb
        metrics = {k: wandb.Html(v) for k, v in metrics.items()}
        wandb.log(metrics, step=step)

    def log_matrix(self, metrics: Dict, step: int, namespace: str):
        wandb = self._hooked.wandb
        metrics = {k: wandb.Image(v) for k, v in metrics.items()}
        wandb.log(metrics, step=step)

    def on_first_exception(self, source: Trainer, func, params: ParamsType, e: BaseException, *args, **kwargs):
        super().on_first_exception(source, func, params, e, *args, **kwargs)


class TensorBoardCallback(RecordCallback):
    only_main_process = True

    def on_hooked(self, source: Trainer, params: ParamsType):
        super().on_hooked(source, params)
        _ = source.safe_writer
        # get experiment property of writer
        args = source.exp.get_prop('tensorboard_args')
        log_dir = args['log_dir']
        source.logger.raw(f'tensorboard --logdir {log_dir}')

    def log_text(self, metrics: Dict, step: int, namespace: str):
        writer = self._hooked.safe_writer
        for k, v in metrics.items():
            writer.add_text(k, v, global_step=step)

    def log_scalars(self, metrics: Dict, step: int, namespace: str):
        writer = self._hooked.safe_writer
        for k, v in metrics.items():
            writer.add_scalar(tag=f'{namespace}.{k}', scalar_value=v,
                              global_step=step)

    def log_matrix(self, metrics: Dict, step: int, namespace: str):
        return NotImplemented


class StopByCode(TrainCallback):
    def __init__(self, step=100):
        self.step = step

    def on_train_step_end(self, trainer: Trainer, func, params: ParamsType, metric: Meter, *args, **kwargs):
        if trainer.global_steps % self.step == 0:
            if os.path.exists(trainer.exp.test_file('.stop')):
                trainer.exp.add_tag('lumo.early_stop')
                trainer.logger.info('Early stop this test manully.')
                trainer.stop_train_epoch()
                trainer.stop_train()


CallbackType = NewType('CallbackType', BaseCallback)


class SeedCallback(InitialCallback):

    def on_imodels_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        super().on_imodels_begin(trainer, func, params, *args, **kwargs)
        seed = params.get('seed', None)
        if seed is None:
            trainer.logger.info('params.seed is None, seed will be set as 0')
            seed = 0
        trainer.rnd.mark(seed)


class RemoteCallback(InitialCallback, TrainCallback):
    only_main_process = True

    def __init__(self, url=None):
        import requests
        self.url = url
        self.req = requests
        self.property = {}
        self.submits = []
        self.executor = ThreadPoolExecutor(1)

    def make_data(self, source, event_name, **experimentInfo):
        return {
            'event': event_name,
            'identity': {
                'test_root': source.exp.test_root,
                'test_name': source.exp.test_name,
                'exp_name': source.exp.exp_name,
                'pid': os.getpid(),
                'ppid': os.getppid(),
            },
            'experimentInfo': experimentInfo,
        }

    def request(self, data: dict):
        task = self.executor.submit(self.req.post, self.url, json={'data': data,
                                                                   'type': 'timeevent',
                                                                   'from': 'lumo.RemoteCallback',
                                                                   'datetime': datetime.now().isoformat()})
        self.submits.append(task)

    def on_hooked(self, source: Trainer, params: ParamsType):
        source.exp.dump_info('request', {
            'url': self.url,
        })
        data = self.make_data(source, 'on_hooked', **{
            'logfile': '\n'.join(source.logger.out_channel),
            'param_hash': params.hash(),
            'gpu_number': world_size(),
            'params': params.to_dict(),
            'args': source.exp.exec_argv,
        })

        self.request(data)

    def on_imodels_end(self, trainer: Trainer, func, params: ParamsType, result: Any, *args, **kwargs):
        super().on_imodels_end(trainer, func, params, result, *args, **kwargs)
        commit_info = trainer.exp.get_prop('git')
        if 'commit' in commit_info:
            self.property['experimentInfo']['commit'] = commit_info['commit']

        data = self.make_data(trainer, 'on_imodels_end', **{
            'commit': commit_info['commit']})
        self.request(data)

    def on_train_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        self.request(self.make_data(trainer, 'on_train_begin'))

    def on_train_end(self, trainer: Trainer, func, params: ParamsType, record: Record, *args, **kwargs):
        self.request(self.make_data(trainer, 'on_train_end'))

    def on_train_epoch_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        self.request(
            self.make_data(trainer, 'on_train_epoch_end', eidx=trainer.eidx, global_steps=trainer.global_steps))

    def finished(self, *args):
        self.request(self.make_data(self._hooked, 'on_finished'))
        _ = [i.result() for i in as_completed(self.submits)]

    def on_train_step_end(self, trainer: Trainer, func, params: ParamsType, metric: MetricType = None, *args,
                          **kwargs):
        super().on_train_step_end(trainer, func, params, metric, *args, **kwargs)
        metric = wrap_result(metric)
        dict(metric.items())
        self.request(self.make_data())

    def on_test_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        self.request(
            self.make_data(trainer, 'on_test_begin', eidx=trainer.eidx, global_steps=trainer.global_steps))

    def on_eval_begin(self, trainer: Trainer, func, params: ParamsType, *args, **kwargs):
        self.request(
            self.make_data(trainer, 'on_eval_begin', eidx=trainer.eidx, global_steps=trainer.global_steps))

    def on_first_exception(self, source: Trainer, func, params: ParamsType, e: BaseException, *args, **kwargs):
        self.request(
            self.make_data(source, 'on_exception',
                           **{
                               'exception': ''.join(traceback.format_exception(type(e), e, e.__traceback__)),
                               'exception_type': type(e)
                           }))



import os
import stat
import subprocess
import sys
from collections import OrderedDict
from pprint import pformat

from joblib import hash

from lumo.decorators.lazy_required import is_lib_available
from lumo.proc.dependency import get_lock
from lumo.proc.dist import is_main
from lumo.utils import safe_io as io
from lumo.utils.exithook import wrap_before
from lumo.utils.fmt import strftime, indent_print
from . import Experiment
from .base import ExpHook as BaseExpHook


class ExpHook(BaseExpHook):
    def regist(self, exp: Experiment):
        self.exp = exp

    def on_start(self, exp: Experiment, *args, **kwargs): pass

    def on_end(self, exp: Experiment, end_code=0, *args, **kwargs): pass

    def on_progress(self, exp: Experiment, step, *args, **kwargs): pass

    def on_newpath(self, exp: Experiment, *args, **kwargs): pass


class LastCmd(ExpHook):
    def on_start(self, exp: Experiment, *args, **kwargs):
        argv = exp.exec_argv
        fn = f'run_{os.path.basename(argv[1])}.sh'

        strings = OrderedDict.fromkeys([i.lstrip('#').strip() for i in io.load_text(fn).split('\n')])

        cur = f"{' '.join(argv)} $@"
        if cur in strings:
            strings.pop(cur)

        with open(fn, 'w', encoding='utf-8') as w:
            w.write('\n'.join([f'# {i}' for i in strings.keys()]))
            w.write('\n\n')
            w.write(cur)

        st = os.stat(fn)
        os.chmod(fn, st.st_mode | stat.S_IEXEC)


class PathRecord(ExpHook):

    def on_newpath(self, exp: Experiment, *args, **kwargs):
        super().on_newpath(exp, *args, **kwargs)


class Diary(ExpHook):
    def on_start(self, exp: Experiment, *args, **kwargs):
        super().on_start(exp, *args, **kwargs)
        with open(exp.root_file(f'{strftime("%y%m%d")}.log', 'diary'), 'a') as w:
            w.write(f'{strftime("%H:%M:%S")}, {exp.test_root}\n')


class RecordAbort(ExpHook):
    def regist(self, exp: Experiment):
        super().regist(exp)
        wrap_before(self.exc_end)

    def exc_end(self, exc_type, exc_val, exc_tb):
        import traceback
        res = traceback.format_exception(exc_type, exc_val, exc_tb)
        res = [i for i in res if 'in _newfunc' not in i]
        self.exp.dump_string('exception', "".join(res))
        self.exp.end(
            end_code=1,
            exc_type=traceback.format_exception_only(exc_type, exc_val)[-1].strip()
        )


class TimeMonitor(ExpHook):
    def _create_agent(self, exp: Experiment):
        from lumo.exp import agent
        cmd = [
            sys.executable, '-m', agent.__spec__.name,
            f"--state_key=state",
            f"--pid={os.getpid()}",
            f"--exp_name={exp.exp_name}",
            f"--test_name={exp.test_name}",
            f"--test_root={exp.test_root}",
            # f"--params={sys.argv}" # TODO add sys.argv
        ]
        subprocess.Popen(' '.join(cmd),
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True,
                         start_new_session=True)

    def on_start(self, exp: Experiment, *args, **kwargs):
        super().on_start(exp)
        self._create_agent(exp)
        exp.dump_info('state', {
            'start': strftime(),
            'end': strftime()
        })


class GitCommit(ExpHook):

    def on_start(self, exp: Experiment, *args, **kwargs):
        if not is_lib_available('git'):
            exp.dump_info('git', {
                'code': -1,
                'msg': 'git not installed'
            })
            return

        from lumo.utils.repository import git_enable, git_commit, git_dir
        from lumo.utils.ast import analyse_module_dependency
        from lumo.utils.hash import hash_iter
        import inspect

        if not git_enable():
            exp.dump_info('git', {
                'code': -1,
                'msg': 'git disabled'
            })
            return

        if not is_main():
            return

        import __main__

        root = git_dir()
        mem = analyse_module_dependency(__main__, root=root)

        filter_files = set()
        dep_source = []
        for fn, dep_module in sorted(mem.items()):
            if os.path.commonprefix([fn, root]) == root:
                filter_files.add(os.path.relpath(fn, root))
                try:
                    dep_source.append(inspect.getsource(dep_module))
                except OSError:
                    pass

        dep_hash = hash_iter(*dep_source)
        commit_ = git_commit(key='lumo', info=exp.test_root, filter_files=filter_files)

        if commit_ is None:
            exp.dump_info('git', {
                'code': -2,
                'msg': 'commit error'
            })
            return

        commit_hex = commit_.hexsha[:8]
        exp.dump_info('git', {
            'commit': commit_hex,
            'repo': exp.project_root,
            'dep_hash': dep_hash,
        })

        file = exp.root_file(hash(exp.project_root), 'repos')
        exps = {}
        if os.path.exists(file):
            exps = io.load_json(file)
        res = exps.setdefault(exp.project_root, list())
        if exp.exp_root not in res:
            res.append(exp.exp_root)
        io.dump_json(exps, file)


class LockFile(ExpHook):

    def on_start(self, exp: Experiment, *args, **kwargs):
        exp.dump_info('lock', get_lock('torch', 'numpy',
                                       'joblib',
                                       'psutil',
                                       'decorator',
                                       'torch',
                                       'numpy',
                                       'accelerate',
                                       'hydra',
                                       'omegaconf', ))


class FinalReport(ExpHook):
    def on_end(self, exp: Experiment, end_code=0, *args, **kwargs):
        # if end_code == 0:
        print('-----------------------------------')
        # print('')

        print('Properties:')
        indent_print(pformat(exp.properties))
        print('Tags:')
        indent_print(pformat(exp.tags))
        print('Use paths:')
        indent_print(pformat(exp.paths))
        print('Execute:')
        indent_print(' '.join(exp.exec_argv))
        print('-----------------------------------')



import numpy as np
import streamlit as st
import sys

sys.path.insert(0, './src/lumo')
from . import parser

from lumo.exp import finder, Experiment
from datetime import date, timedelta


class HeadType:
    manully = 'Manully Type'
    query = 'Filter & Search'


class Main():
    def __init__(self):
        test_names = self.select_head()
        for test_name in test_names:
            test_root = finder.ensure_test_root(test_name)
            if test_root:
                # with st.expander(test_root):
                self.make_test(test_root)
                # st.write(test_root)
                # st.write(finder.format_experiment(Experiment.from_disk(test_root)))
            else:

                st.error(f'"{test_name}" is not a valid test name/ test root!')

    def make_test(self, test_root: str):
        exp = Experiment.from_disk(test_root)
        with st.expander("Experiment Info"):
            st.write(finder.format_experiment(exp))
        # with st.expander("Visualize Metrics"):
        if exp.has_prop('tensorboard_args'):
            tb = exp.get_prop('tensorboard_args')
            metrics = parser.parse_fron_tensorboard(tb['log_dir'])
        elif exp.has_prop('logger_args'):
            tb = exp.get_prop('logger_args')
            metrics = parser.parse_from_log(tb['log_dir'])
        else:
            metrics = {}
        metrics = list(metrics.items())
        for i in range(0, len(metrics), 2):
            l, m = st.columns(2)
            k, v = metrics[i]
            l.write(k)
            l.line_chart(np.array([vv.value for vv in v]))
            if i + 1 >= len(metrics):
                break
            k, v = metrics[i + 1]
            m.write(k)
            m.line_chart(np.array([vv.value for vv in v]))
                # if i + 2 >= len(metrics):
                #     break
                # k, v = metrics[i + 2]
                # r.line_chart({'k': np.array([vv.value for vv in v])})

    def select_head(self):
        left, right = st.columns([1, 3])
        left, right = st.sidebar, st.sidebar
        # left = st
        res = left.selectbox('', [HeadType.manully, HeadType.query])
        if res == HeadType.manully:
            test_names = self.manully_head(right)
        else:
            test_names = self.query_head(right)
        return test_names

    def manully_head(self, st):
        c = 0
        c = st.number_input('Test Number', c)
        test_names = []
        for i in range(c):
            test_names.append(st.text_input(f'Test ({i})', help='type test name or test root'))

        return test_names

    def query_head(self, st):
        sd = st
        self.experiments = finder.find_experiments()
        default_exp_name = None

        selection = sd.multiselect('Experiments', self.experiments, default=default_exp_name)

        test_names = []
        for select in selection:
            for test_name in finder.list_test_names_from_experiment(select):
                test_names.append(f'{select}/{test_name}')

        use_date_filter = False
        use_date_filter = sd.checkbox('Use Date Filter', use_date_filter)

        start = sd.date_input('Date Start Filter', date.today() - timedelta(days=7), disabled=not use_date_filter)
        end = sd.date_input('Date End Filter', date.today(), disabled=not use_date_filter)

        def unwrap(exp_test_name):
            return exp_test_name.split('/')[1]

        if use_date_filter:
            left = start.strftime('%y%m%d')
            right = end.strftime('%y%m%d')
            test_names = [i for i in test_names if left <= unwrap(i).split('.')[0] <= right]

        test_names = sd.multiselect('Tests', test_names)
        return [unwrap(i) for i in test_names]


def main():
    Main()


if __name__ == '__main__':
    main()



import warnings
from collections import OrderedDict
from pprint import pformat
from typing import NewType, Union

from torch.utils.data import DataLoader

from lumo.core.metaclasses import PropVar

__all__ = ['DataLoader', 'LumoDataLoader', 'DataLoaderSide', 'DataLoaderType']


class DataLoaderIterWrap:
    def __init__(self, iter_fn, batch_count=None):
        self.iter_fn = iter_fn
        self.iter = iter_fn()
        self.c = 0
        self.batch_count = batch_count

    def __iter__(self):
        while True:
            try:
                yield next(self)
            except StopIteration:
                break

    def __len__(self):
        if self.batch_count is None:
            return len(self.iter)
        else:
            return self.batch_count

    def __next__(self):
        if self.batch_count is not None:
            if self.c >= self.batch_count:
                raise StopIteration()
        try:
            batch = next(self.iter)
        except StopIteration as e:
            if self.batch_count is not None:
                self.iter = self.iter_fn()
                batch = next(self.iter)
            else:
                raise e

        self.c += 1
        return batch


class LumoDataLoader(DataLoader):

    def _wraooed_iter_(self):
        return DataLoaderIterWrap(super().__iter__, len(self))

    def __iter__(self) -> DataLoaderIterWrap:
        return self._wraooed_iter_()


def summarize_loader(loader: DataLoader):
    if isinstance(loader, DataLoaderSide):
        inner = pformat({f"{k}(cycle={loader._cycle[k]})": summarize_loader(v) for k, v in loader._loaders.items()})
        return f"DataLoaderSide({inner})"
    elif isinstance(loader, DataLoader):
        size = '?'
        try:
            size = len(loader)
        except:
            pass
        if loader.batch_sampler is not None:
            batch_size = loader.batch_sampler.batch_size

        # clss = type(loader).__mro__
        # cls_str = []
        # for cls in clss:
        #     if isinstance(cls, DataLoader):
        #         if len(cls_str) == 0:
        #             cls_str.append(cls.__name__)
        #         break
        #     else:
        #         cls_str.append(cls.__name__)
        #
        # cls_str = '|'.join(cls_str)
        return f"{loader.__class__.__name__}(batch_size={batch_size}, num_workers={loader.num_workers}, size={size})"
    else:
        raise ValueError(f'Need {DataLoaderType}, got type {type(loader)}')


class DataLoaderSide:
    """
    `DataLoaderSide` is used when different DataLoader with different batch_size are feeded at the same time.
    """

    def __init__(self):
        self._loaders = OrderedDict()
        self._cycle = OrderedDict()
        self._state = 'zip'

    @property
    def dataset(self):
        return {k: v.dataset for k, v in self.source.items()}

    @property
    def source(self):
        return self._loaders

    def add(self, name, loader: DataLoader, cycle=False):
        self._loaders[name] = loader
        self._cycle[name] = cycle
        return self

    def copy(self):
        loader = DataLoaderSide()
        loader._loaders = self._loaders
        loader._cycle = self._cycle
        loader._state = self._state
        return loader

    def zip(self):
        self._state = 'zip'
        return self

    def chain(self):
        self._state = 'chain'
        return self

    def __len__(self):
        valid_keys = [k for k, cycle in self._cycle.items() if not cycle]
        return min([len(self._loaders[k]) for k in valid_keys])

    def __iter__(self):
        iters = {k: iter(v)
                 for k, v in self._loaders.items()}
        stop = None
        while stop is None:
            res = OrderedDict()
            for k, v in iters.items():
                try:
                    batch = next(v)
                except StopIteration as e:
                    if self._cycle[k]:
                        v = iter(self._loaders[k])
                        iters[k] = v
                        batch = next(v)
                    else:
                        # stop = e
                        return
                res[k] = batch
            if self._state == 'zip':
                yield res
            else:
                yield list(res.values())


DataLoaderType = NewType('DataLoaderType', Union[DataLoader, DataLoaderSide])



from .params import *
from .metaclasses import *
from .attr import Attr
from . import interp
from .meter import Meter
from .record import Record, wrap_result, MetricType
from .enums import TrainStage



"""

"""
__version__ = "0.14.3"

from .core import Params, ParamsType, MetricType, Meter, Record, TrainStage, BaseParams
from .data import DataLoader, DataModule, DatasetBuilder, LumoDataLoader, CollateBase, DataLoaderSide
from .exp import SimpleExperiment, Experiment
from .trainer import Trainer, TrainerParams, callbacks, RndManager
from .utils import Logger



"""
Methods about screen single line outputs.
"""
import os
import shutil
import sys
import time

from tqdm import tqdm


def get_consolo_width():
    return shutil.get_terminal_size().columns - 1  # -1 for windows consolo


def support_multiline():
    if "jupyter_core" in sys.modules or shutil.get_terminal_size((0, 0)).columns == 0 or "PYCHARM_HOSTED" in os.environ:
        return True
    else:
        return False


def _is_jupyter() -> bool:  # pragma: no cover
    """Check if we're running in a Jupyter notebook."""
    try:
        get_ipython  # type: ignore
    except NameError:
        return False
    ipython = get_ipython()  # type: ignore
    shell = ipython.__class__.__name__
    if "google.colab" in str(ipython.__class__) or shell == "ZMQInteractiveShell":
        return True  # Jupyter notebook or qtconsole
    elif shell == "TerminalInteractiveShell":
        return False  # Terminal running IPython
    else:
        return False  # Other type (?)


class ScreenStr:
    """
    A ScreenStr start with '\r' won't overflow, any string outside the screen width will be cut.

    Notes:
    If output consolo support multiline(like pycharm or jupyter notebook) return, all string will be represented.
    """
    t = 0
    dt = 0.7
    last = 0
    left = 0
    max_wait = 1.
    wait = 0
    wait_toggle = False

    debug = False
    last_width = 0
    multi_mode = support_multiline()

    def __init__(self, content="", leftoffset=0) -> None:
        self.content = content
        ScreenStr.left = leftoffset

    def __repr__(self) -> str:
        if ScreenStr.multi_mode:
            return self.content
        return self._screen_str()

    def tostr(self):
        return self.content

    @classmethod
    def set_speed(cls, dt: float = 0.05):
        cls.dt = dt

    @classmethod
    def deltatime(cls):
        if cls.last == 0:
            cls.last = time.time()
            return 0
        else:
            end = time.time()
            res = end - cls.last
            cls.last = end
            return res

    @classmethod
    def cacu_offset_(cls, out_width):

        delta = cls.deltatime()
        cls.t += delta * cls.dt

        # pi = 2*math.pi
        t = cls.t
        # k = 2 * out_width / pi
        k = 10
        pi = 2 * out_width / k
        offset = round(k * (t % pi) * ((t % pi) < pi / 2) + (-k * (t % pi) + 2 * out_width) * ((t % pi) > pi / 2))

        # offset = math.floor(out_width * (math.cos(ScreenStr.t + math.pi) + 1) / 2)
        # print(offset)
        return offset

    a = 1

    def __len__(self) -> int:
        txt = self.content.encode("gbk", errors='ignore')
        return len(txt)

    def _decode_sub(self, txt, left, right):
        try:
            txt = txt[left:right].decode("gbk", errors='ignore')
        except:
            try:
                txt = txt[left:right - 1].decode("gbk", errors='ignore')
            except:
                try:
                    txt = txt[left + 1:right].decode("gbk", errors='ignore')
                except:
                    txt = txt[left + 1:right - 1].decode("gbk", errors='ignore')

        return txt

    @staticmethod
    def consolo_width():
        width = get_consolo_width()
        return width

    @staticmethod
    def split(txt, len):
        try:
            return txt[:len], txt[len:]
        except:
            try:
                return txt[:len + 1], txt[len + 1:]
            except:
                return txt[:len - 1], txt[len - 1:]

    def _screen_str(self, margin="..."):
        width = self.consolo_width()

        txt = self.content.encode("gbk", errors='ignore').strip()
        textlen = len(txt)

        if textlen <= width:
            return self.content

        left, right = self.split(txt, self.left)
        if len(left) >= width:
            return left[:width]

        offset = 0

        offright = width - len(left) + offset - len(margin)

        left = left.decode("gbk", errors='ignore')
        right = self._decode_sub(right, offset, offright)

        head = "\r" if self.content.startswith("\r") else ""
        tail = "\n" if self.content.endswith("\n") else ""

        txt = "{}{}{}{}".format(head, left, right, tail)
        return txt + margin


class inlinetqdm(tqdm):

    def full_str(self):
        return self.format_meter(**self.format_dict)

    def __str__(self):
        return ScreenStr(self.full_str())._screen_str()



import enum


class TrainStage(enum.Enum):
    default = 'default'
    train = 'train'
    test = 'test'
    val = 'val'

    def is_train(self):
        return self.value == 'train'

    def is_test(self):
        return self.value == 'test'

    def is_val(self):
        return self.value == 'val'

    @staticmethod
    def create_from_str(value):
        if value in {'eval', 'evaluate'}:
            value = 'val'
        return TrainStage(value)



import warnings
from numbers import Number

from . import Attr
from .metaclasses import PropVar
from .meter import Meter
import torch
import numpy as np
from typing import NewType, Mapping, Union, Sequence, Dict
from collections import OrderedDict

MetricType = NewType('MetricType', Union[Mapping, Meter, Sequence,
                                         Number, torch.Tensor, np.ndarray
])

from collections import namedtuple

record_event = namedtuple('record_event', ['global_step', 'metric', 'time'])


def wrap_result(metric: MetricType) -> Meter:
    """
    Wrap any form of metric data into Meter form.

    """
    if isinstance(metric, (Meter,)):
        return metric
    elif isinstance(metric, Mapping):
        return Meter.from_dict(metric)
    elif isinstance(metric, Sequence):
        return Meter.from_dict({f'm{i}': v for i, v in enumerate(metric)})
    elif isinstance(metric, (torch.Tensor, np.ndarray, float, int)):
        return Meter.from_dict({'metric': metric})
    return Meter()


class Record(metaclass=PropVar):
    def __init__(self, window_size=500, **kwargs):
        self._prop.update(kwargs)
        self._cache = []
        self._agg = OrderedDict()  # type:Dict[str,AggItem]

    def avg(self) -> Attr:
        warnings.warn('avg will be deprecated in later version, use agg() instead.')
        return self.agg()

    def agg(self) -> Attr:
        res = Attr()
        for k, v in self._agg.items():
            res[k] = v.res
        return res

    @property
    def stage(self):
        return self._prop['stage']

    def __str__(self):
        res = self.agg()
        rep = []
        for k, v in res.items():
            if isinstance(v, float):
                rep.append(f'{k}={v:.4g}')
            else:
                rep.append(f'{k}={str(v)}')

        return ', '.join(rep)

    def tostr(self):
        return str(self)

    def record(self, metric, global_step=None):
        meter = wrap_result(metric)
        agg = meter._avg

        for k, v in meter.items():
            stg = agg.get(k, 'last')
            item = self._agg.get(k, None)
            if item is None:
                item = AggItem(stg)
            item.update(v)
            self._agg[k] = item

    def clear(self):
        self._agg.clear()
        self._cache.clear()

    def flush(self):
        self._cache.clear()


class AggItem:
    def __init__(self, stg):
        self.stg = stg
        self._last = 0
        self.acc = 0
        self.c = 0

    @property
    def res(self):
        if self.stg == 'mean':
            return self.acc / self.c

        if self.stg in {'min', 'max', 'last'}:
            return self.acc

        if self.stg == 'sum':
            return self.acc

        return self.acc

    @property
    def last(self):
        return self._last

    def update(self, val):
        if self.stg == 'last':
            self.acc = val
        elif self.stg == 'min':
            self.acc = min(self.acc, val)
        elif self.stg == 'max':
            self.acc = max(self.acc, val)
        elif self.stg in {'mean', 'sum'}:
            self.acc += val
            self.c += 1
        self._last = val



"""
Methods about git.
"""
import os
import warnings
from functools import lru_cache

import git
from git import Repo, Commit
import io
from joblib import hash
from .filelock2 import Lock

LUMO_BRANCH = 'lumo_experiments'

_commits_map = {}


class branch:
    """
    用于配合上下文管理切换 git branch

    with branch(repo, branch):
        repo.index.commit('...')
    """

    def __init__(self, repo: Repo, branch: str):
        self.repo = repo
        self.lock = Lock(f'{hash(repo.git_dir)}_{branch}')
        self.lock.abtain()
        self.old_branch = self.repo.head.reference
        self.branch = branch

    def __enter__(self):
        if self.branch is None:
            return

        if self.branch not in self.repo.heads:
            head = self.repo.create_head(self.branch)
        else:
            head = self.repo.heads[self.branch]

        self.repo.head.set_reference(head)
        # self.repo.head.reset(index=True, working_tree=True)
        return head

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.branch is None:
            return
        self.repo.head.set_reference(self.old_branch)
        self.lock.release()


def check_have_commit(repo):
    if len(repo.heads) == 0:
        repo.git.add('.')
        repo.index.commit('initial commit')


@lru_cache()
def load_repo(dir='./'):
    """
    Try to load git repository object of a directory.
    Args:
        dir: str, a directory path, default is the current working dir.
        if dir is a repository dir, then a git.Repo object will be retured.
        if not, some you can type a path to init it, or type '!' to cancel init it.

    Returns:
        git.Repo object or None if dir not have git repository and cancel to init it.
    """
    path = git_dir(dir)
    repo = Repo(path)
    check_have_commit(repo)
    return repo


def add(repo=None):
    if repo is None:
        repo = load_repo()
    return repo.git.add(all=True)


def git_commit(repo=None, key=None, branch_name=LUMO_BRANCH, info: str = None, filter_files=None):
    """
    ```
        cd <repo working dir>
        git reset <branch_name>
        git add .
        git commit -m "<info>"
        git reset <original branch>
    ```
    Args:
        repo:
        key:
            to avoid duplicate commit, a key value can be passed,
            commit operation will be perform if `key` hasn't appeared before.
            default value is None, means you can commit as you want without limitation
        branch_name:
            commit on which branch, nonexistent branch will be created.
        info:
            commit info, a string
    Returns:
        git.Commit object, see gitpython for details.
    """
    try:
        if repo is None:
            repo = load_repo()

        if key is not None and key in _commits_map:
            return _commits_map[key]

        if LUMO_BRANCH not in repo.branches:
            repo.create_head(LUMO_BRANCH)
            print(f'branch {LUMO_BRANCH} not found, will be created automatically.')

        exp_head_commit = repo.heads[LUMO_BRANCH].commit
        diff = repo.active_branch.commit.diff(exp_head_commit)

        if filter_files is not None:
            diff = [i.a_path for i in diff if i.a_path in filter_files]

        if len(diff) == 0:
            commit_ = exp_head_commit
        else:
            with branch(repo, branch_name):
                change_file = []
                change_file.extend(repo.untracked_files)
                change_file.extend([i.a_path for i in repo.head.commit.diff(None)])
                # print(change_file)
                if filter_files is not None:
                    print('before filter', change_file)
                    change_file = [i for i in change_file if i in filter_files]
                print('after filter', change_file)

                repo.git.add(change_file)
                commit_info = '[[EMPTY]]'
                if info is not None:
                    commit_info = info
                commit_ = repo.index.commit(commit_info)

        if key is not None:
            _commits_map[key] = commit_
    except (git.GitCommandError, ValueError, IndexError) as e:
        commit_ = None
        print(f'git commit Exception {e}')
    return commit_


def reset(repo=None, commit_hex=None, commit: Commit = None):
    """
    将工作目录中的文件恢复到某个commit
    恢复快照的 git 流程:
        git branch experiment
        git add . & git commit -m ... // 保证文件最新，防止冲突报错，这一步由 Experiment() 代为完成
        git checkout <commit-id> // 恢复文件到 <commit-id>
        git checkout -b reset // 将当前状态附到新的临时分支 reset 上
        git branch experiment // 切换回 experiment 分支
        git add . & git commit -m ... // 将当前状态重新提交到最新
            // 此时experiment 中最新的commit 为恢复的<commit-id>
        git branch -D reset  // 删除临时分支
        git branch master // 最终回到原来分支，保证除文件变动外git状态完好
    Returns:
        An Experiment represents this reset operation
    """
    if repo is None:
        repo = load_repo()

    if commit is None and commit_hex is not None:
        commit = repo.commit(commit_hex)

    old_path = os.getcwd()
    os.chdir(commit.tree.abspath)

    with branch(commit.repo, LUMO_BRANCH) as new_branch:
        repo.git.checkout(commit.hexsha)
        repo.git.checkout('-b', 'reset')
        repo.head.reference = new_branch
        _ = git_commit(repo, branch_name=repo.head.reference.name, info="Reset from {}".format(commit.hexsha))
        repo.git.branch('-d', 'reset')

    os.chdir(old_path)
    return None


def archive(repo=None, commit_hex=None, commit: Commit = None, tgt=None):
    """
    TODO
    将某次 test 对应 commit 的文件打包，相关命令为
        git archive -o <filename> <commit-hash>
    Returns:
        An Experiment represents this archive operation
    """
    if repo is None:
        repo = load_repo()

    old_path = os.getcwd()
    os.chdir(commit.tree.abspath)
    # exp = Experiment('Archive')

    # revert_path = checkpath(cache_dir(), 'archives', commit)
    # revert_fn = os.path.join(revert_path, "code.zip")

    # TODO 在code.zip目录下添加相关说明
    # exp.add_plugin('archive', {'file': revert_fn,
    #                            'test_name': self.name})
    # with open(revert_fn, 'wb') as w:
    #     repo.archive(w, commit)

    # exp.end()
    os.chdir(old_path)
    return None


@lru_cache(1)
def git_enable():
    try:
        import git
    except ImportError:
        warnings.warn('python library `gitpython` not installed, git operations will be ignored. '
                      'If you want lumo to use git to manage your code, please install it by executing `pip install gitpython`')
        return False

    try:
        git.Git().execute(['git', 'rev-parse', '--git-dir'])
        return True
    except git.GitCommandError:
        return False


def git_dir(root='./'):
    """
    git repository directory
    git rev-parse --git-dir
    Args:
        root:
    Returns:

    """
    if git_enable():
        from git import Git
        cur = os.getcwd()
        os.chdir(root)
        res = Git().execute(['git', 'rev-parse', '--git-dir'])
        res = os.path.abspath(os.path.dirname(res))
        os.chdir(cur)
        return res
    else:
        return None


def get_tree_from_commit(commit: Commit, tree=None):
    if tree is None:
        tree = commit.tree
    yield tree.abspath, tree.blobs, tree.trees
    for tree in tree.trees:
        yield from get_tree_from_commit(commit, tree)


def get_diff_tree_from_commits():
    pass


def get_file_of_commit(commit: Commit, file_name) -> bytes:
    blob = commit.tree / file_name
    return blob.data_stream.read()



import numpy as np


def mixup(major, minor=None, lam=None, alpha=2):
    if lam is None:
        lam = beta(alpha)
    lam = max(lam, 1 - lam)
    if minor is None:
        reid = permutation(len(major))
        minor = major[reid]
    return major * lam + minor * (1 - lam), lam


def permutation(size, nonoverlap=True):
    if nonoverlap:
        half = size // 2
        reid = np.concatenate(
            [
                np.random.permutation(half) + half,
                np.random.permutation(half),
            ]
        )
    else:
        reid = np.random.permutation(size)
    return reid


def beta(alpha=2.0):
    """
    Return lambda
    Args:
        alpha:

    Returns:

    """
    if alpha > 0.:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1.
    return lam



import os
import sys
from .config import HOME, EXP_ROOT, CACHE_ROOT, BLOB_ROOT


def home():
    if HOME:
        return HOME
    return os.path.expanduser("~")


def cache_dir():
    try:
        os.makedirs(CACHE_ROOT, exist_ok=True)
        res = CACHE_ROOT
    except PermissionError:
        res = os.path.join(home(), '.lumo/.cache')
        os.makedirs(res, exist_ok=True)
    return res


def dataset_cache_dir(name=None):
    if name is None:
        res = os.path.join(cache_dir(), 'datasets', '__default__')
    else:
        res = os.path.join(cache_dir(), 'datasets', name)
    os.makedirs(res, exist_ok=True)
    return res


def libhome():
    return os.path.join(home(), '.lumo')


def exproot():
    if EXP_ROOT:
        return EXP_ROOT
    return libhome()


def blobroot():
    if BLOB_ROOT:
        return BLOB_ROOT
    return libhome()


def local_dir():
    from lumo.utils.repository import git_dir
    res = git_dir()
    if res is None:
        res = os.path.dirname(os.path.join(os.getcwd(), sys.argv[0]))
    return res



from dbrecord import PDict
import pandas as pd


def collect(*database):
    res = []
    for exp in database:
        exp = PDict(exp)
        for v in exp.values():
            if isinstance(v, dict):
                res.append(v)
        exp.close()

    return pd.DataFrame(res)



from torch.optim import Optimizer
import torch


class LARS(Optimizer):
    """
    Layer-wise adaptive rate scaling
    - Converted from Tensorflow to Pytorch from:
    https://github.com/google-research/simclr/blob/master/lars_optimizer.py
    - Based on:
    https://github.com/noahgolmant/pytorch-lars
    params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): base learning rate (\gamma_0)
        lr (int): Length / Number of layers we want to apply weight decay, else do not compute
        momentum (float, optional): momentum factor (default: 0.9)
        use_nesterov (bool, optional): flag to use nesterov momentum (default: False)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0.0)
            ("\beta")
        eta (float, optional): LARS coefficient (default: 0.001)
    - Based on Algorithm 1 of the following paper by You, Gitman, and Ginsburg.
    - Large Batch Training of Convolutional Networks:
        https://arxiv.org/abs/1708.03888
    """

    def __init__(self, params, lr, len_reduced, momentum=0.9, use_nesterov=False, weight_decay=0.0,
                 classic_momentum=True, eta=0.001):

        self.epoch = 0
        defaults = dict(
            lr=lr,
            momentum=momentum,
            use_nesterov=use_nesterov,
            weight_decay=weight_decay,
            classic_momentum=classic_momentum,
            eta=eta,
            len_reduced=len_reduced
        )

        super(LARS, self).__init__(params, defaults)
        self.lr = lr
        self.momentum = momentum
        self.weight_decay = weight_decay
        self.use_nesterov = use_nesterov
        self.classic_momentum = classic_momentum
        self.eta = eta
        self.len_reduced = len_reduced

    def step(self, epoch=None, closure=None):

        loss = None

        if closure is not None:
            loss = closure()

        if epoch is None:
            epoch = self.epoch
            self.epoch += 1

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            eta = group['eta']
            learning_rate = group['lr']

            # TODO: Hacky
            counter = 0
            for p in group['params']:
                if p.grad is None:
                    continue

                param = p.data
                grad = p.grad.data

                param_state = self.state[p]

                # TODO: This really hacky way needs to be improved.
                # Note Excluded are passed at the end of the list to are ignored
                if counter < self.len_reduced:
                    grad += self.weight_decay * param

                # Create parameter for the momentum
                if "momentum_var" not in param_state:
                    next_v = param_state["momentum_var"] = torch.zeros_like(
                        p.data
                    )
                else:
                    next_v = param_state["momentum_var"]

                if self.classic_momentum:
                    trust_ratio = 1.0

                    # TODO: implementation of layer adaptation
                    w_norm = torch.norm(param)
                    g_norm = torch.norm(grad)

                    device = g_norm.get_device()

                    trust_ratio = torch.where(w_norm.ge(0), torch.where(
                        g_norm.ge(0), (self.eta * w_norm / g_norm), torch.Tensor([1.0]).to(device)),
                                              torch.Tensor([1.0]).to(device)).item()

                    scaled_lr = learning_rate * trust_ratio

                    next_v.mul_(momentum).add_(scaled_lr, grad)

                    if self.use_nesterov:
                        update = (self.momentum * next_v) + (scaled_lr * grad)
                    else:
                        update = next_v

                    p.data.add_(-update)

                # Not classic_momentum
                else:

                    next_v.mul_(momentum).add_(grad)

                    if self.use_nesterov:
                        update = (self.momentum * next_v) + (grad)

                    else:
                        update = next_v

                    trust_ratio = 1.0

                    # TODO: implementation of layer adaptation
                    w_norm = torch.norm(param)
                    v_norm = torch.norm(update)

                    device = v_norm.get_device()

                    trust_ratio = torch.where(w_norm.ge(0), torch.where(
                        v_norm.ge(0), (self.eta * w_norm / v_norm), torch.Tensor([1.0]).to(device)),
                                              torch.Tensor([1.0]).to(device)).item()

                    scaled_lr = learning_rate * trust_ratio

                    p.data.add_(-scaled_lr * update)

                counter += 1

        return loss



"""

"""
from typing import Any, Mapping, Sequence
from lumo.core.params import ParamsType
import numpy as np
import torch
from torch.utils.data._utils.collate import default_collate


class CollateBase:

    @classmethod
    def from_collate(cls, collate_fn, params: ParamsType = None):
        instance = cls(params)
        instance._collate_fn = collate_fn
        return instance

    def __init__(self, params: ParamsType = None) -> None:
        super().__init__()
        self._collate_fn = default_collate
        self.params = params

    def __call__(self, *args, **kwargs):
        res = self.before_collate(*args, **kwargs)
        res = self._collate_fn(res)
        res = self.after_collate(res)
        return res

    def before_collate(self, sample_list):
        return sample_list

    def raw_collate(self, sample_list):
        return self._collate_fn(sample_list)

    def collate(self, sample_list):
        return self._collate_fn(sample_list)

    def after_collate(self, batch):
        return batch


class IgnoreNoneCollate(CollateBase):

    def _filter_none(self, item):
        if item is None:
            return False
        if isinstance(item, (list, tuple)):
            return all([self._filter_none(i) for i in item])
        if isinstance(item, dict):
            return all(self._filter_none(i) for i in item.values())
        return True

    def before_collate(self, sample_list):
        return list(filter(self._filter_none, sample_list))


def numpy_collate(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""

    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        batch = [elem.detach().cpu().numpy() for elem in batch]
        return np.stack(batch, 0)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            return numpy_collate([np.array(b) for b in batch])
        elif elem.shape == ():  # scalars
            return np.array(batch)
    elif isinstance(elem, float):
        return np.array(batch, dtype=np.float)
    elif isinstance(elem, int):
        return np.array(batch)
    elif isinstance(elem, (str, bytes)):
        return batch
    elif isinstance(elem, Mapping):
        return {key: numpy_collate([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(numpy_collate(samples) for samples in zip(*batch)))
    elif isinstance(elem, Sequence):
        # check to make sure that the elements in batch have consistent size
        it = iter(batch)
        elem_size = len(next(it))
        if not all(len(elem) == elem_size for elem in it):
            raise RuntimeError('each element in list of batch should be of equal size')
        transposed = zip(*batch)
        return [numpy_collate(samples) for samples in transposed]



import os.path
from dbrecord import PList
from lumo.proc import path
from lumo.utils.filelock2 import Lock
from lumo.utils import safe_io as IO


class Metrics:
    """
    Record metrics at multiple steps. Supported by dbrecord.
    """

    def __init__(self, test_path: str):
        os.makedirs(test_path, exist_ok=True)
        self.fpath = os.path.join(test_path, f'metric_board.sqlite')
        self.disk = PList(self.fpath)
        self.lock = Lock(os.path.basename(test_path.rstrip('/')))

    def append(self, metric: dict, step, stage='train'):
        self.disk.append({
            'metric': metric,
            'step': step,
            'stage': stage
        })
        self.disk.flush()

    def flush(self):
        self.disk.flush()


class TableRow:
    """
    It can be regarded as a serialized dictionary,
    or a certain row in the table, so the same key value will be overwritten.

    If you need to record records at different times, please use trainer.metrics
    """

    def __init__(self, table, partition, rowkey):
        dirpath = os.path.join(path.libhome(), 'metrics', table)
        os.makedirs(dirpath, exist_ok=True)
        self.fpath = os.path.join(dirpath, partition, f'{rowkey}.pkl')
        self.key = rowkey
        self.value = {}
        # self.disk = PDict(self.fpath)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.flush()

    def flush(self):
        IO.dump_pkl(self.value, self.fpath)

    def update_metrics(self, dic: dict, compare=None, flush=False):
        res = {}
        [res.update(self.update_metric(k, v, compare)) for k, v in dic.items()]

        if flush:
            self.flush()
        return res

    def update_metric(self, key, value, compare=None, flush=False):
        dic = self.metric
        old = dic.setdefault(key, None)

        update = False
        if old is None or compare is None:
            update = True
        else:
            if compare == 'max':
                if old < value:
                    update = True
            elif compare == 'min':
                if old > value:
                    update = True
            else:
                assert False

        if update:
            dic[key] = value
            old = value

        if flush:
            self.flush()

        return {key: value}

    @property
    def metric(self):
        return self.value.setdefault('metric', {})

    def update_metric_pair(self, key, value, key2, value2, compare=None, flush=False):
        dic = self.metric
        old = dic.setdefault(key, None)
        old2 = dic.setdefault(key2, None)

        update = False
        if old is None or compare is None:
            update = True
        else:
            if compare == 'max':
                if old < value:
                    update = True
            elif compare == 'min':
                if old > value:
                    update = True
            else:
                assert False

        if update:
            dic[key] = value
            dic[key2] = value2
            old, old2 = value, value2

        if flush:
            self.flush()

        return {key: value, key2: value2}

    def set_params(self, params: dict):
        self.value['params'] = params
        self.flush()
        return params

    def update_dict(self, dic: dict, flush=False):
        for k, v in dic.items():
            self.update(k, v)
        if flush:
            self.flush()

    def update(self, key, value, flush=True):
        self.value[key] = value
        if flush:
            self.flush()



"""
heartbeat mechanism
"""
import time

import psutil
from joblib import hash

from lumo.core import BaseParams
from lumo.utils import safe_io as IO
from lumo.utils.fmt import strftime
from lumo.exp import Experiment


def wait_pid_stop(exp_name=None, test_root=None, state_key='state1'):
    params = BaseParams()
    params.pid = None
    params.exp_name = exp_name
    params.state_key = state_key
    params.from_args()

    exp = Experiment.from_disk(test_root)

    test_root = exp.test_root
    while params.pid is None or psutil.pid_exists(params.pid):
        exp.dump_info(params.state_key, {
            'end': strftime(),
        }, append=True)

        IO.dump_text(test_root, exp.root_branch.file(f'{hash(test_root)}.hb', 'heartbeat'))

        time.sleep(2)


if __name__ == '__main__':
    wait_pid_stop()



"""
Longest common subsequence
edited from http://wordaligned.org/articles/longest-common-subsequence
"""
import itertools
from difflib import Match, SequenceMatcher
from collections import namedtuple
from lumo.contrib.itertools import window

Chunk = namedtuple('Chunk', 'w a size')


class LCS:
    """
    A diiflib.SequenceMatch like class.

    Examples:
        a, b = 'GTCGTTCGGAATGCCGTTGCTCTGTAAA', 'ACCGGTCGAGTGCGCGGAAGCCGGCCGAA'
        l = LCS(a, b)
        print(l)
        for m in l.get_matching_blocks():
            print(m, a[m.a:m.a + m.size], b[m.b:m.b + m.size])
    """

    def __init__(self, xstr, ystr):
        self.xstr = xstr
        self.ystr = ystr

    def get_matching_blocks(self):
        """
        Return list of triples describing matching subsequences.
        """
        res = self.lcs2()
        if len(res) == 0:
            return []

        li, lj = res[0][1], res[0][3]
        size = 1
        ress = []
        for x, i, y, j in res[1:]:
            offset = (i - li) * (j - lj)
            if offset == 1:
                li, lj = i, j
                size += 1
            else:
                end = size - 1
                ress.append(Match(li - end, lj - end, size))
                li, lj = i, j
                size = 1
        end = size - 1
        ress.append(Match(li - end, lj - end, size))
        return ress

    def get_matching_chunks(self):
        """
        Return list of triples describing matching subsequences.
        """
        res = self.lcs2()
        if len(res) == 0:
            return []

        li, lj = res[0][1], res[0][3]
        size = 1
        chunk_x = []
        chunk_y = []
        for x, i, y, j in res[1:]:
            offset = (i - li) * (j - lj)
            if offset == 1:
                li, lj = i, j
                size += 1
            else:
                end = size - 1
                chunk_x.append([self.xstr[li - end:li - end + size], li - end, size])
                chunk_y.append([self.ystr[lj - end:lj - end + size], lj - end, size])
                li, lj = i, j
                size = 1
        end = size - 1
        chunk_x.append([self.xstr[li - end:li - end + size], li - end, size])
        chunk_y.append([self.ystr[lj - end:lj - end + size], lj - end, size])

        # a, b = self.xstr, self.ystr

        def _merge(ress, a):
            for i in range(len(ress)):
                m = ress[i]
                l = ress[i - 1] if i > 0 else []
                r = ress[i + 1] if i < len(ress) - 1 else []
                if len(m) > 0 and m[-1] == 1:
                    if len(l) > 0 and a[l[1] + l[2]] == m[0]:
                        l[0] = l[0] + m[0]
                        l[2] += 1
                        m.clear()
                    elif len(r) > 0 and a[r[1] - 1] == m[0]:
                        r[0] = m[0] + r[0]
                        r[1] -= 1
                        r[2] += 1
                        m.clear()
            ress = [Chunk(i[0], i[1], i[2]) for i in ress if len(i) > 0]
            return ress

        chunk_x = _merge(chunk_x, self.xstr)
        chunk_y = _merge(chunk_y, self.ystr)
        return chunk_x, chunk_y

    def lcs(self):
        return lcs(self.xstr, self.ystr)

    def lcs2(self):
        return lcs2(self.xstr, self.ystr)


def lcs_lens(xs, ys):
    curr = list(itertools.repeat(0, 1 + len(ys)))
    for x in xs:
        prev = list(curr)
        for i, y in enumerate(ys):
            if x == y:
                curr[i + 1] = prev[i] + 1
            else:
                curr[i + 1] = max(curr[i], prev[i + 1])
    return curr


def lcs(xs, ys):
    """
    Return longest common subsequence list.
    """
    nx, ny = len(xs), len(ys)
    if nx == 0:
        return []
    elif nx == 1:
        return [xs[0]] if xs[0] in ys else []
    else:
        i = nx // 2
        xb, xe = xs[:i], xs[i:]
        ll_b = lcs_lens(xb, ys)
        ll_e = lcs_lens(xe[::-1], ys[::-1])
        _, k = max((ll_b[j] + ll_e[ny - j], j)
                   for j in range(ny + 1))
        yb, ye = ys[:k], ys[k:]
        return lcs(xb, yb) + lcs(xe, ye)


def lcs2(xs, ys, ii=0, kk=0):
    """
    Return longest common subsequence list with each character's index.
    """
    nx, ny = len(xs), len(ys)
    if nx == 0:
        return [['', -1, '', -1]]
    elif nx == 1:
        idx = ys.find(xs[0])
        if idx >= 0:
            return [[xs[0], ii, ys[idx], idx + kk]]
        return []
    else:
        i = nx // 2
        xb, xe = xs[:i], xs[i:]
        ll_b = lcs_lens(xb, ys)
        ll_e = lcs_lens(xe[::-1], ys[::-1])
        _, k = max((ll_b[j] + ll_e[ny - j], j)
                   for j in range(ny + 1))
        yb, ye = ys[:k], ys[k:]
        return lcs2(xb, yb, ii, kk) + lcs2(xe, ye, i + ii, k + kk)


import numpy as np


def lcs3(text1, text2, ii=0, kk=0):
    m, n = len(text1), len(text2)
    dp = [[0] * (n + 1) for _ in range(m + 1)]
    mask = [[0] * (n + 1) for _ in range(m + 1)]

    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if text1[i - 1] == text2[j - 1]:
                dp[i][j] = dp[i - 1][j - 1] + 1
                mask[i][j] = 0
            else:
                dp[i][j] = max(dp[i - 1][j], dp[i][j - 1])
                mask[i][j] = 2 * int(dp[i - 1][j] > dp[i][j - 1]) - 1

    i, j = m, n
    match = []
    while i != 0 and j != 0:
        if mask[i][j] == 0:
            match.insert(0, [text1[i - 1], i - 1, text2[j - 1], j - 1])
            i, j = i - 1, j - 1
        elif mask[i][j] == 1:
            i, j = i - 1, j
        else:
            i, j = i, j - 1
    return match



import copy
import warnings
from functools import partial
from pprint import pformat

from typing import Callable, NewType, Dict, Any, Iterable, Sequence

import numpy as np
from torch.utils.data import Dataset

from lumo.data.loader import LumoDataLoader

SingleValueTransform = NewType('SingleValueTransform', Callable[[Any], Any])
DictTransform = NewType('DictTransform', Callable[[Dict[str, Any]], Any])


class DatasetBuilder(Dataset):
    def __init__(self):
        self._prop = {}

        self._idx_keys = set()
        self._data = {}
        self._outs = {}
        self._transforms = {}

        self._outkeys = []

        self._iter_cache = {}

    def __repr__(self):

        if self.sized:
            return f'Builder(flow={pformat(self._outs)}, sized={self.sized}, size={len(self)}, iterable={self.iterable})'
        else:
            return f'Builder(flow={pformat(self._outs)}, sized={self.sized}, iterable={self.iterable})'

    def __iter__(self):
        if self.subindices is not None:
            warnings.warn('subset() has no effects in iter mode.')

        for key, outkeys in self._outs.items():
            if key == '::idx::':
                warnings.warn(f'iter does not support idx, will skip {outkeys}')
                continue

            source = self._data[key]
            for outkey in outkeys:
                self._iter_cache[outkey] = iter(source)
        return self

    def __next__(self):
        if len(self._iter_cache) == 0:
            raise StopIteration()
        try:
            outputs = {k: next(v) for k, v in self._iter_cache.items()}
            if self.mode != 'zip':
                outputs = [outputs[outkey] for outkey in self._outkeys]
            return outputs
        except StopIteration as e:
            self._iter_cache.clear()
            raise e

    def __getitem__(self, index):

        index = self.map_index(index)

        outputs = {}

        for key, outkeys in self._outs.items():
            if key == '::idx::':
                ipt = index
            else:
                ipt = self._data[key][index]
            ipt_transform = self._transforms.get(key, None)
            if ipt_transform is not None:
                ipt = ipt_transform(ipt)

            for outkey in outkeys:
                opt = ipt
                opt_transform = self._transforms.get(f'::{outkey}', None)
                if opt_transform is not None:
                    opt = opt_transform(opt)

                outputs[outkey] = opt

        glb_transform = self._transforms.get('::global::', None)
        if glb_transform is not None:
            outputs = glb_transform(outputs)

        if self.mode == 'chain':
            outputs = [outputs[outkey] for outkey in self._outkeys]
        elif self.mode == 'item':
            outputs = list(outputs.values())[0]
        return outputs

    def __len__(self):
        if not self.sized:
            raise TypeError('DatasetBuilder can not be sized when data source is empty or only iterable.')

        return self._prop['__clen__']

    def _update_len(self):
        if not self.sized:
            self._prop['__clen__'] = None
            return

        if self.pseudo_length is not None:
            res = self.pseudo_length
            self._prop['__clen__'] = res
            return res

        if self.subindices is not None:
            res = len(self.subindices)
        else:
            res = self._prop['__len__']

        if self.pseudo_repeat is not None:
            res = res * self.pseudo_repeat
        self._prop['__clen__'] = res
        return res

    @property
    def inputs(self):
        return self._data

    @property
    def outputs(self):
        mapping = {}
        for key, outkeys in self._outs.items():
            if key == '::idx::':
                source = range(len(self))
            else:
                source = self._data[key]

            for outkey in outkeys:
                mapping[outkey] = source
        return mapping

    @property
    def mode(self):
        return self._prop.get('mode', 'zip')

    @property
    def iterable(self):
        return self._prop.get('iterable', False)

    @property
    def subindices(self):
        return self._prop.get('subindices', None)

    @property
    def pseudo_length(self) -> int:
        return self._prop.get('pseudo_length', None)

    @property
    def pseudo_repeat(self) -> int:
        return self._prop.get('pseudo_repeat', None)

    def copy(self):
        db = DatasetBuilder()
        db._prop = copy.copy(self._prop)
        db._idx_keys = copy.copy(self._idx_keys)
        db._data = copy.copy(self._data)
        db._outs = copy.deepcopy(self._outs)
        db._transforms = copy.copy(self._transforms)
        db._outkeys = copy.copy(self._outkeys)
        return db

    def subset(self, indices: Sequence[int]):
        self._prop['subindices'] = np.array(indices)
        self._update_len()
        return self

    def scale_to_size(self, size: int):
        assert isinstance(size, int)
        assert 'pseudo_repeat' not in self._prop
        assert 'pseudo_length' not in self._prop

        self._prop['pseudo_length'] = size
        self._update_len()
        return self

    def repeat(self, multiple: int):
        assert isinstance(multiple, int)
        assert 'pseudo_length' not in self._prop
        assert 'pseudo_repeat' not in self._prop

        self._prop['pseudo_repeat'] = multiple
        self._update_len()
        return self

    def map_index(self, index):
        """
        Map the raw index to the final index for source data.
        Args:
            index:

        Returns:

        """
        if self.pseudo_length is not None or self.pseudo_repeat is not None:
            if self.subindices is not None:
                index = index % len(self.subindices)
                index = self.subindices[index]
            else:
                index = index % self._prop['__len__']
        elif self.subindices is not None:
            index = self.subindices[index]

        return index

    @property
    def sized(self):
        return self._prop.get('sized', False)

    def chain(self):
        self._prop['mode'] = 'chain'
        return self

    def item(self):
        self._prop['mode'] = 'item'
        return self

    def zip(self):
        self._prop['mode'] = 'zip'
        return self

    def _check_source(self, name, source):
        # source is sized can be itered
        # source can be itered not meant it is sizable.
        if self.subindices is not None:
            return ValueError('subset() should be called after preparing all data source.')
        if self.pseudo_repeat is not None or self.pseudo_length is not None:
            return ValueError('scale_to_size() or repeat() should be called after preparing all data source.')

        if self._prop.get('sized', True):
            size = self._prop.get('__len__', None)
            try:
                newsize = len(source)
                if size is None:
                    self._prop['sized'] = True
                    self._prop['iterable'] = True
                    self._prop['__len__'] = newsize
                    self._update_len()
                elif size != newsize:
                    warnings.warn(f'Source {name} has different length from '
                                  f'other registed source.')

            except (TypeError, NotImplementedError):
                self._prop['sized'] = False
                self._prop['__len__'] = None

        if isinstance(source, Iterable):
            self._prop['iterable'] = True

        if not self._prop.get('iterable', False) and not self._prop.get('sized', False):
            raise TypeError(f'Source {name} must be an iterable or sized object, but got {type(source)}.')

    def add_idx(self, name):
        outkeys = self._outs.setdefault(f"::idx::", [])
        assert name not in self._outkeys, f'Output key {name} duplicated.'
        outkeys.append(name)
        self._outkeys.append(name)
        return self

    def add_input(self, name: str, source, transform: SingleValueTransform = None):
        assert name not in self._data, f'Source name {name} duplicated.'
        self._check_source(name, source)
        self._data[name] = source
        self._transforms[name] = transform
        return self

    def add_input_transform(self, name: str, transform: SingleValueTransform = None):
        assert name in self._data, f'Source {name} should be added.'
        self._transforms[name] = transform
        return self

    def add_output(self, name: str, outkey: str, transform: SingleValueTransform = None):
        assert name in self._data, f'Must have data source {name} first.'

        outkeys = self._outs.setdefault(name, list())
        assert outkey not in self._outkeys, f'Output key {outkey} duplicated.'
        outkeys.append(outkey)
        self._outkeys.append(outkey)

        self._transforms[f'::{outkey}'] = transform
        return self

    def add_output_transform(self, outkey: str, transform: SingleValueTransform = None):
        assert outkey in self._outkeys, f'Output key {outkey} should be added.'
        self._transforms[f'::{outkey}'] = transform
        return self

    def add_global_transform(self, transform: DictTransform):
        self._transforms['::global::'] = transform
        return self

    def set_input_transform(self, name, transform: SingleValueTransform):
        self._transforms[name] = transform
        return self

    def set_output_transform(self, outkey, transform: SingleValueTransform):
        self._transforms[f'::{outkey}'] = transform
        return self

    DataLoader = LumoDataLoader
    DL = LumoDataLoader  # alias to DataLoader

    def __getattribute__(self, item):
        res = super(DatasetBuilder, self).__getattribute__(item)
        if item == 'DataLoader':
            res = partial(res, dataset=self)
        return res

    def get_source(self, name):
        return self._data[name]



"""
Format date/filename, check array shape, convert item from torch.Tensor to ndarray or scalar.
"""
import textwrap
from datetime import datetime

import numpy as np
import torch

from . import re


def to_ndarray(item):
    if isinstance(item, torch.Tensor):
        item = item.detach().cpu()
    return np.array(item)


def detach(item):
    if isinstance(item, torch.Tensor):
        item = item.detach().cpu().numpy()
    return item


def validate_scalar_shape(ndarray, name=''):
    if ndarray.ndim != 0:
        raise ValueError(
            "Expected scalar value for %r but got %r" % (name, ndarray)
        )
    return ndarray


def is_scalar(ndarray: np.ndarray):
    return ndarray.size == 1


def strftime(fmt='%y-%m-%d-%H%M%S', dateobj: datetime = None):
    """get current date with formatted"""
    if dateobj is not None:
        return dateobj.strftime(fmt)
    return datetime.now().strftime(fmt)


_invalid_fc = (
    r"[+?@#$&%*()=;|,<>: +"
    r"\^\-\/\t\b\[\]\"]+"
)


def to_filename(basename):
    return re.sub(_invalid_fc, '_', basename)


def can_be_filename(basename):
    return re.search(_invalid_fc, basename) is None


def indent_print(text, indent='    '):
    print(textwrap.indent(text, indent))



from lumo.exp import Experiment
import os

import zipfile


def test_extract(test_root, output=None, verbose=True):
    exp = Experiment.from_disk(test_root)
    if output is None:
        output = os.path.join(os.getcwd(), f'{exp.test_name}.zip')

    z = zipfile.ZipFile(output, 'a', zipfile.ZIP_DEFLATED)

    if verbose:
        print('deflate info')
    for root, dirs, fs in os.walk(exp.test_root):
        for f in fs:
            a = os.path.join(root, f)
            b = os.path.join('/', exp.exp_name, exp.test_name, root.replace(exp.test_root, 'experiment').lstrip('/'), f)
            z.write(a, b
                    )
            if verbose:
                print(f'{a} => {b}')

    if verbose:
        print('deflate blob')
    for root, dirs, fs in os.walk(exp.blob_root):
        for f in fs:
            a = os.path.join(root, f)
            b = os.path.join('/', exp.exp_name, exp.test_name, root.replace(exp.blob_root, 'blob/'), f)
            z.write(a, b
                    )
            if verbose:
                print(f'{a} => {b}')



from lumo.core.interp import Linear, Cos, PeriodCos, PeriodLinear, PeriodHalfCos


def test_continous():
    assert 0.5 == Linear.interp(5, start=0, end=1, left=0, right=10)
    assert 0.5 == Linear(start=0, end=1, left=0, right=10)(5)
    assert 0.5 == Cos.interp(5, start=0, end=1, left=0, right=10)
    assert 0.5 == Cos(start=0, end=1, left=0, right=10)(5)

    assert 0 == Cos(start=0, end=1, left=0, right=10, constant=True)(5)


def test_period():
    assert 0.5 == PeriodLinear.interp(15, start=0, end=1, left=0, period=10)
    assert 0.5 == PeriodHalfCos.interp(15, start=0, end=1, left=0, period=10)



import os
import warnings

CAN_USE_LOCK = True
if os.name == 'nt':
    import win32con, win32file, pywintypes

    LOCK_EX = win32con.LOCKFILE_EXCLUSIVE_LOCK
    LOCK_SH = 0  # The default value
    LOCK_NB = win32con.LOCKFILE_FAIL_IMMEDIATELY
    __overlapped = pywintypes.OVERLAPPED()


    def lock(file, flags):
        hfile = win32file._get_osfhandle(file.fileno())
        win32file.LockFileEx(hfile, flags, 0, 0xffff0000, __overlapped)


    def unlock(file):
        hfile = win32file._get_osfhandle(file.fileno())
        win32file.UnlockFileEx(hfile, 0, 0xffff0000, __overlapped)
elif os.name == 'posix':
    import fcntl


    def lock(file, flags=fcntl.LOCK_EX):
        fcntl.flock(file.fileno(), flags)


    def unlock(file):
        fcntl.flock(file.fileno(), fcntl.LOCK_UN)
else:
    CAN_USE_LOCK = False

    warnings.warn(f'You are in an unknown platform {os.name}, filelock may cant be used.')


    def lock(file, flat):
        raise NotImplementedError(f'an UNKNOWN platform {os.name}')


    def unlock(file):
        fcntl.flock(file.fileno(), fcntl.LOCK_UN)



"""
This module aims to provide an easy and general method to generate series values, which can be used in
 applying learning rate of optimizer or scaling a loss value as weight.

To meet this demand, the concept 'schedule' is summarized as a periodic math function, which have
left/right interval and  start/end value.

Hence, a `Schedule` class is provided, it receives four base parameters: start, end, left, and right corresponding to
the above concept respectively.

This class provides some common methods. For example, when you have a schedule instance, you can apply learning rate by
simple call `schedule.scale()` or `schedule.apply()` function.

And, you can use `schedule.plot()` to plot a curve of the values in each step. The plot function use `matplotlib` as backend.

If you don't want to create an instance, you can call classmethod `get(cls, cur, start=0, end=1, left=0, right=1)` to get
value.

Except for the base `Schedule` class, some other subclasses of `Schedule`  which may be general used is provided, too. All can
 be easily understand by their names and plot curves.
"""
from typing import List

import numpy as np

from lumo.core import BaseParams
from lumo.core.metaclasses import PropVar

__all__ = ['Cos',
           'Linear',
           'Exp',
           'Log',
           'Constant',
           'PeriodCos',
           'PeriodHalfCos',
           'PeriodTriangle',
           'PeriodLinear',
           'PowerDecay',
           'InterpolateList', ]


class Interpolate(BaseParams):
    """
    ratio 变化为从 1 - 0
    """

    def toggle_constant(self, toggle=True):
        """fix the schedule as the first value"""
        self.constant = toggle
        return self

    @classmethod
    def interp(self, *args, **kwargs):
        raise NotImplementedError()

    def __repr__(self):
        content = ', '.join(["{}={}".format(k, v) for k, v in self.items()])
        return "{}({})".format(self.__class__.__name__, content)

    def __call__(self, cur):
        raise NotImplementedError()

    def plot(self, num=1000, left=0, right=1000, show=True):
        """
        Plot a curve of the schedule.
        Args:
            num:
            left:
            right:

        Returns:
            plt.plot

        Notes:
            You may need to call `plt.show`() to show it.
        """
        from matplotlib import pyplot as plt

        x = np.linspace(left, right, num)
        y = [self(i) for i in x]

        res = plt.plot(x, y)
        plt.title(str(self))
        if show:
            plt.show()
        return res

    def scale(self, optimizer, cur):
        """
        Scale the learning rate by current value. 'scale' means not apply the current schedule value directly to
        the learning rate, but multiple the initial learning rate. You can use `schedule.apply()` to apply the schedule
        value directly.

        Notes:
        -------
        When first apply scale function, a `_raw_lr` represent initial lr will be set in each param_group, then, the
        learning rate(store in param_groups with key 'lr') will be calculated by `_raw_lr * schedule(cur)`.

        Args:
            optimizer: A pytorch optimizer instance.
            cur: current step of this schedule.

        Returns:
            Current schedule value.
        """
        ratio = self(cur)
        for param_group in optimizer.param_groups:  # type:dict
            raw_lr = param_group.setdefault('_raw_lr', param_group['lr'])
            param_group['lr'] = raw_lr * ratio

        return ratio

    def apply(self, optimizer, cur):
        """
        Apply the learning rate with current schedule value.

        Args:
            optimizer: A pytorch optimizer instance.
            cur: current step of this schedule.

        Returns:

        """
        new_lr = self(cur)
        for param_group in optimizer.param_groups:  # type:dict
            param_group['lr'] = new_lr

        return new_lr


class ABCContinuous(Interpolate):

    def __call__(self, cur):
        if self.constant:
            return self.start
        return self.interp(cur, start=self.start, end=self.end, left=self.left, right=self.right,
                           constant=self.constant)

    def __init__(self, start=1e-3, end=1e-6, left=0, right=80, *args, **kwargs):
        super().__init__()
        self.left = left
        self.right = right
        self.start = start
        self.end = end
        self.constant = kwargs.get('constant', (left == right))

    @classmethod
    def ratio(cls, cur, left, right, constant=False):
        if constant:
            return 0
        return (cur - left) / (right - left)

    @classmethod
    def get_val(cls, cur, start=1e-3, end=1e-6, left=0, right=80, *args, **kwargs):
        """get the current schedule value without create `schedule` instance. """
        return cls.interp(cur=cur, start=start, end=end, left=left, right=right, *args, **kwargs)

    def plot(self, num=1000, left=None, right=None, show=True):
        if left is None:
            left = self.left

        if right is None:
            right = self.right

        if self.constant:
            right = left + 50

        return super().plot(num, left, right, show)


class ABCPeriod(Interpolate):
    """
    period
    """

    def __init__(self, start=0, end=1, left=0, period=1, *args, **kwargs):
        super().__init__()
        self.left = left
        self.period = period
        self.start = start
        self.end = end
        self.constant = (period == 0)

    @classmethod
    def ratio(cls, cur, left, period, constant=False):
        if constant:
            return 0
        if cur < left:
            in_period = (period - (left - cur) % period) % period
        else:
            in_period = float(cur - left) % period
        return in_period / period

    @classmethod
    def get_val(cls, cur, start=0, end=1, left=0, right=1, *args, **kwargs):
        """get the current schedule value without create `schedule` instance. """
        return cls.interp(cur=cur, start=start, end=end,
                          left=left, right=right, *args, **kwargs)

    def plot(self, num=1000, left=None, n_period=5, show=True):
        if left is None:
            left = self.left

        right = self.period * n_period
        if self.constant:
            right = left + 50

        return super().plot(num, left, right, show)

    def __call__(self, cur):
        return self.interp(cur, start=self.start, end=self.end, left=self.left, period=self.period,
                           constant=self.constant)


class Cos(ABCContinuous):
    """one cycle cosine functoin"""

    @classmethod
    def interp(cls, cur, start=0., end=1., left=0., right=1., *args, **kwargs):
        constant = kwargs.get('constant', False)
        if constant:
            return start

        if cur < left:
            return start
        elif cur > right:
            return end

        ratio = cls.ratio(cur, left=left, right=right, constant=constant)
        cos_ratio = 0.5 * (1 + np.cos(ratio * np.pi))
        return start * cos_ratio + end * (1 - cos_ratio)

    def get(self, cur):
        return self(cur)


class Linear(ABCContinuous):
    """linear schedule"""

    @classmethod
    def interp(cls, cur, start=0., end=1., left=0., right=1., *args, **kwargs):
        constant = kwargs.get('constant', False)
        if constant:
            return start

        if cur < left:
            return start
        elif cur > right:
            return end

        linear_ratio = cls.ratio(cur, left=left, right=right, constant=constant)
        return start * (1 - linear_ratio) + end * linear_ratio


class Exp(ABCContinuous):
    """slow to quick"""

    @classmethod
    def interp(cls, cur, start=0., end=1., left=0., right=1., *args, **kwargs):
        constant = kwargs.get('constant', False)
        if constant:
            return start

        if cur < left:
            return start
        elif cur > right:
            return end

        ratio = cls.ratio(cur, left=left, right=right, constant=constant)
        residual = np.exp(-5)

        exp_ratio = np.exp((ratio - 1) * 5) - residual * (1 - ratio)
        return start * (1 - exp_ratio) + end * exp_ratio


class Log(ABCContinuous):
    """quick to slow"""

    @classmethod
    def interp(cls, cur, start=0., end=1., left=0., right=1., *args, **kwargs):
        constant = kwargs.get('constant', False)
        if constant:
            return start

        if cur < left:
            return start
        elif cur > right:
            return end

        ratio = cls.ratio(cur, left=left, right=right, constant=constant)

        residual = np.exp(-5)

        log_ratio = 1 - np.exp(-ratio * 5) + residual * ratio
        return start * (1 - log_ratio) + end * log_ratio


class Constant(ABCContinuous):
    def __init__(self, value=0.5, *args, **kwargs):
        super().__init__(start=value, end=value, left=0, right=1, *args, **kwargs)
        self.constant = True


class PeriodCos(ABCPeriod):
    """
    periodic cosine schedule
    """

    @classmethod
    def interp(cls, cur, start=0., end=1., left=0., period=1., *args, **kwargs):
        constant = kwargs.get('constant', False)

        ratio = cls.ratio(cur, left=left, period=period, constant=constant)
        cos_ratio = 0.5 * (1 + np.cos(ratio * np.pi * 2))
        return start * cos_ratio + end * (1 - cos_ratio)


class PeriodHalfCos(ABCPeriod):
    """
    half periodic cosine schedule, period is (right-left)
    """

    @classmethod
    def interp(cls, cur, start=0., end=1., left=0., period=1., *args, **kwargs):
        constant = kwargs.get('constant', False)
        ratio = cls.ratio(cur, left=left, period=period, constant=constant)
        cos_ratio = 0.5 * (1 + np.cos(ratio * np.pi))
        return start * cos_ratio + end * (1 - cos_ratio)


class PeriodTriangle(ABCPeriod):
    def __init__(self, start=0, end=1, left=0, left_period=1, right_period=1, *args, **kwargs):
        super().__init__(start=start, end=end, left=left,
                         period=(left_period + right_period),
                         *args, **kwargs)
        assert left_period > 0 and right_period > 0
        self.left_period = left_period
        self.right_period = right_period

    @classmethod
    def interp(cls, cur, start=0., end=1., left=0., left_period=0., right_period=1., *args, **kwargs):
        constant = kwargs.get('constant', False)
        ratio = cls.ratio(cur, left=left, period=(left_period + right_period), constant=constant)

        mid_ratio = left_period / (right_period + left_period)

        if ratio < mid_ratio:
            ratio = ratio / mid_ratio
            return start * (1 - ratio) + end * ratio
        else:
            ratio = (ratio - mid_ratio) / (1 - mid_ratio)
            return end * (1 - ratio) + start * ratio


class PeriodLinear(ABCPeriod):
    """
    sawtooth wave, like a period line schedule
    """

    @classmethod
    def interp(cls, cur, start=0., end=1., left=0., period=1., *args, **kwargs):
        constant = kwargs.get('constant', False)
        ratio = cls.ratio(cur, left=left, period=period, constant=constant)
        return start * (1 - ratio) + end * ratio


class PowerDecay(Interpolate):
    """equal to tf.train.exponential_decay, decay every <decay_steps> with a base of <decay_rate> """

    def __init__(self, start, decay_steps, decay_rate, end=None):
        super().__init__()
        self.start = start
        self.decay_steps = decay_steps
        self.decay_rate = decay_rate
        self.end = end

    def __call__(self, cur):
        rate = np.power(self.decay_rate, cur // self.decay_steps)
        res = self.start * rate
        if self.end is not None:
            res = max(self.end, res)

        return res


class PowerDecay2(Interpolate):
    def __init__(self, start, schedules, gammas):
        super().__init__()
        self.start = start
        self.schedules = schedules
        self.gammas = gammas

    @classmethod
    def interp(cls, cur, start=0., gammas=None, schedules=None, *args, **kwargs):
        if schedules is None:
            schedules = []
        if gammas is None:
            gammas = []

        res = start
        for (gamma, step) in zip(gammas, schedules):
            if (cur >= step):
                res = res * gamma
            else:
                break
        return res

    def __call__(self, cur):
        self.interp(cur, start=self.start, gammas=self.gammas, schedules=self.schedules)


class InterpolateList(Interpolate):
    def __init__(self, schedules: List[Interpolate]):
        super().__init__()
        self.schedules = schedules
        if len(schedules) > 1:
            for i in range(len(schedules) - 1):
                if schedules[i].right < schedules[i + 1].left:
                    raise ValueError('schedule[i].right should equal or larger than schedule[i+1].left, '
                                     f'but got {schedules[i].right} and {schedules[i + 1].left}')

            self.left = self.schedules[0].left
            self.right = self.schedules[-1].right

    def __reduce__(self):
        return (self.__class__, (self.schedules, self.bound))

    def __call__(self, cur):
        for i, schedule in enumerate(self.schedules):
            if i + 1 < len(self.schedules):
                if cur < self.schedules[i + 1].left:
                    return schedule(cur)
                else:
                    continue
            else:
                return schedule(cur)

    def __repr__(self):
        content = ', '.join([i.__class__.__name__ for i in self.schedules])
        return '{}({})'.format(self.__class__.__name__, content)

    def plot(self, num=1000, left=None, right=None, show=True):
        if left is None:
            left = self.left

        if right is None:
            right = self.right

        return super().plot(num, left, right, show)



import time

import os
import random

from lumo.utils.exithook import wrap_before


class Lock:
    def __init__(self, name, sleep=1):
        from lumo.proc.path import cache_dir
        self.file = os.path.join(cache_dir(), f"LUMO_LOCK_{name}")
        self.sleep = sleep
        wrap_before(self.clear)

    def clear(self, *_, **__):
        self.release()

    def abtain(self):
        mulp = 1
        while True:
            mulp += 1
            if mulp > 10:
                raise TimeoutError(f'Can not abtain resource of {self.file}')

            while os.path.exists(self.file):
                time.sleep(random.randint(mulp, mulp ** 2))
                mulp += 1
                if mulp > 10:
                    raise TimeoutError(f'Can not abtain resource of {self.file}')

            while True:
                flag = f'{os.getpid()}'
                with open(self.file, 'w') as w:
                    w.write(flag)

                if os.path.exists(self.file):
                    with open(self.file, 'r') as r:
                        lock_flag = r.read()
                        if flag == lock_flag:
                            return True
                        mulp += 1
                        time.sleep(random.randint(mulp, mulp ** 2))
                    if mulp > 10:
                        raise TimeoutError(f'Can not abtain resource of {self.file}')

    def release(self):
        try:
            os.remove(self.file)
        except:
            pass
        return True

    def __enter__(self):
        self.abtain()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()



class ExpHook:

    def regist(self, exp): self.exp = exp

    def on_start(self, exp, *args, **kwargs): pass

    def on_end(self, exp, end_code=0, *args, **kwargs): pass

    def on_progress(self, exp, step, *args, **kwargs): pass

    def on_newpath(self, exp, *args, **kwargs): pass



import git
from pathlib import Path

from lumo.utils.repository import git_commit

git_ignore = ['# Byte-compiled / optimized / DLL files', '.DS_Store', '__pycache__/', '*.py[cod]', '*$py.class', '',
              '# C extensions', '*.so', '', '# Distribution / packaging', '.Python', 'build/', 'develop-eggs/', 'dist/',
              'downloads/', 'eggs/', '.eggs/', 'lib/', 'lib64/', 'parts/', 'sdist/', 'var/', 'wheels/',
              'share/python-wheels/',
              '*.egg-info/', '.installed.cfg', '*.egg', 'MANIFEST', '', '# PyInstaller',
              '#  Usually these files are written by a python script from a template',
              '#  before PyInstaller builds the exe, so as to inject date/other infos into it.', '*.manifest', '*.spec',
              '',
              '# Installer logs', 'pip-log.txt', 'pip-delete-this-directory.txt', '', '# Unit test / coverage reports',
              'htmlcov/',
              '.tox/', '.nox/', '.coverage', '.coverage.*', '.cache', 'nosetests.xml', 'coverage.xml', '*.cover',
              '*.py,cover',
              '.hypothesis/', '.pytest_cache/', 'cover/', '', '# Translations', '*.mo', '*.pot', '', '# Django stuff:',
              '*.log',
              'local_settings.py', 'db.sqlite3', 'db.sqlite3-journal', '', '# Flask stuff:', 'instance/',
              '.webassets-cache', '',
              '# Scrapy stuff:', '.scrapy', '', '# Sphinx documentation', 'docs/_build/', '', '# PyBuilder',
              '.pybuilder/',
              'target/', '', '# Jupyter Notebook', '.ipynb_checkpoints', '', '# IPython', 'profile_default/',
              'ipython_config.py',
              '', '# pyenv', '#   For a library or package, you might want to ignore these files since the code is',
              '#   intended to run in multiple environments; otherwise, check them in:', '# .python-version', '',
              '# pipenv',
              '#   According to pypa/pipenv#598, it is recommended to include Pipfile.lock in version control.',
              '#   However, in case of collaboration, if having platform-specific dependencies or dependencies',
              "#   having no cross-platform support, pipenv may install dependencies that don't work, or not",
              '#   install all needed dependencies.', '#Pipfile.lock', '',
              '# PEP 582; used by e.g. github.com/David-OConnor/pyflow',
              '__pypackages__/', '', '# Celery stuff', 'celerybeat-schedule', 'celerybeat.pid', '',
              '# SageMath parsed files',
              '*.sage.py', '', '# Environments', '.env', '.venv', 'env/', 'venv/', 'ENV/', 'env.bak/', 'venv.bak/', '',
              '# Spyder project settings', '.spyderproject', '.spyproject', '', '# Rope project settings',
              '.ropeproject', '',
              '# mkdocs documentation', '/site', '# pycharm', '.idea', '', '', '# mypy', '.mypy_cache/', '.dmypy.json',
              'dmypy.json',
              '', '# Pyre type checker', '.pyre/', '', '# pytype static type analyzer', '.pytype/', '',
              '# Cython debug symbols',
              'cython_debug/', '.thexp/', 'repo.json', '.expsdirs', '.idea/', '*.pth', '*.npy',
              '*.ckpt',
              '*.thexp.*', '*.pkl', '.cache/', '.lumo/config.json', '*.lumo.*', '*.ft', 'kk', 'temp', '.idea', '',
              'lumo_temp',
              '.lumo/', '*scratch*']


def check_gitignore(path):
    ignore_file = Path(path).joinpath('.gitignore')
    if ignore_file.exists():
        res = ignore_file.read_text().split('\n')
        new_ignore = [i for i in git_ignore if i not in res]
    else:
        new_ignore = git_ignore

    Path(path).joinpath('.gitignore').write_text('\n'.join(new_ignore))


def git_init(path=None):
    if path is None:
        path = '/'

    repo = git.Repo.init(path, mkdir=True)
    check_gitignore(path)
    git_commit(repo, branch_name=None)
    return repo
